# __init__.py
__version__ = '1.0.0'
__author__ = 'Meng Li'

import insitu_alignment as AL
import insitu_DP as DP
import insitu_IO as IO
import insitu_Preprocess as PP
import insitu_Diff as Diff
import insitu_EMraw as EM
import insitu_DENS as DENS


__all__ = ['function1', 'Class2']



# -*- coding: utf-8 -*-
"""
in-situ TEM Toolbox - Movie alignment 

Assembles of functions related to movie alignment

Example:
    
    import insitu_alignment as Al
    Al.get_matrix()

Created on Tue Jun 11 17:22:45 2019
By: Meng Li 
"""


import numpy as np
import csv
import cv2
from skimage import io #package for TIFF stack reading
import tifffile
from scipy import signal
import tqdm
import matplotlib.pyplot as plt
from insituTEM import insitu_IO as IO


def get_path():
    from tkinter import Tk
    from tkinter.filedialog import askopenfilename #Tkinter will provide UI for file selection
    '''
    Retrieves a path via gui
    '''
    Tk().withdraw() 
    # we don't want a full GUI, so keep the root window from appearing
    path = askopenfilename() 
    # show an "Open" dialog box and return the path to the selected file
    return path

def get_matrix(csvpath = None):
    '''
    Reads a csvfile and pulls the dx and dy data
    '''
    if csvpath == None:    
        csvpath = get_path()
        print('''
=============================================================================
Please choose the csv you want to use
              ''')
    with open(csvpath) as csvfile:
        print("\nreading csv\n")
        readcsv = csv.reader(csvfile, delimiter=',')
        dx = [] 
        dy = []
        for row in readcsv:
            dx.append(float(row[1]))
            dy.append(float(row[2]))
        #Currently only works if the csv is saved without headers
    return dx, dy, csvpath


def Adaptive_template_match(path, NOF = 50, csvout=True, TMout=True, TMdenoise=True, TMdenoise_window=5,extend=True, bg = 150):
    '''
    Function to produce adaptive template matching on tiff image stack based on ROI selection
    input: 
        path: image path for grayscale Tiff stack
        NOF: number of frames for template to grow
        csvout: whether to write the translation matrix
        TMout: whether to write the template match found feature
        TMdenoise: whether to denoise the measured matrix to remove sudden change
        TMdenoise_window: window size for TMdenoise
        bg: background color (0-255)
        extend: whether to extend the image for the translation (avoid crop)
    
    output: 
        Translated  image stack
        Csv file of translation matrix
        Captured ROI location

    Write by Meng Li (mona.mengli@gmail.com) 2024.02.06
    '''
    import cv2
    import numpy as np
    from matplotlib import pyplot as plt
    import tifffile
    import easygui
    import tqdm
    from insituTEM import insitu_IO as IO
    # from insituTEM import insitu_alignment as AL
 
    ###############################################################
    ## 1. Inital settings 

    stack = tifffile.imread(path)
    nFrames, h,w = stack.shape
    
    TM = np.zeros((nFrames,3))#frame, TX,TY,max_val
    pathout3 = path[:-4]+'_TM.csv'
    pathout2 = path[:-4]+'_TM.tif'
    
    ################################################################
    #2. Open fr0 and select ROI for template matching
    img = stack[0]
    ## Select ROI for template matching
    r = cv2.selectROI("Select ROI, press ENTER to confirm",img,fromCenter=False,showCrosshair=False)
    cv2.destroyWindow("Select ROI, press ENTER to confirm")
    # Crop image
    template = img[int(r[1]):int(r[1]+r[3]), int(r[0]):int(r[0]+r[2])]#opencv coordinate: Y,x , from upper left corner of the image 
    # Display cropped image
    cv2.imshow("Template, press ENTER to confirm", template)
    cv2.waitKey(0)
    # cv2.destroyWindow("Template, press ENTER to confirm")
    cv2.destroyAllWindows()
    for _ in range(5):
        cv2.waitKey(1)
    
    
    
    # ################################################################
    #3. Apply template to template matching
    meth='cv2.TM_CCOEFF_NORMED'
    method = eval(meth)
    w, h = template.shape[::-1]
 
    print("------------------------------------------")
    print("Calculating drift~")  
    with tifffile.TiffWriter(pathout2,bigtiff=True) as tif:
        for i in tqdm.tqdm(range(nFrames)):
            fr=stack[i]
            # Apply template Matching
            res = cv2.matchTemplate(fr,template,method)
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
            top_left = max_loc
            TM[i,0]=i
            TM[i,1]=top_left[0]
            TM[i,2]=top_left[1]
            bottom_right = (top_left[0] + w, top_left[1] + h)
            cv2.rectangle(fr,top_left, bottom_right, 255, 2)

            if i%NOF==NOF-1:
                template = fr[int(top_left[1]):int(top_left[1] + h), int(top_left[0]):int(top_left[0] + w)]#opencv coordinate: Y,x , from upper left corner of the image 
            if TMout==True: 
                tif.write(fr, contiguous=True)

        
    TM[:,1]=TM[0,1]-TM[:,1]
    TM[:,2]=TM[0,2]-TM[:,2]
    
    # ################################################################
    #4. Apply TM to movie
    if TMdenoise==True: 
        #Get average to minimize total movement
        TM=denoiseTM(TM,TMdenoise_window)
        Mx=int(np.mean(TM[:,1]))
        My=int(np.mean(TM[:,2]))
        TM[:,1]=TM[:,1]-Mx
        TM[:,2]=TM[:,2]-My
    
    print("------------------------------------------")
    print("Saving files~")   
    plt.plot(TM[:,0],TM[:,1],'b',TM[:,0],TM[:,2],'r')

    if csvout==True:
        IO.writeCSV(pathout3,TM,fmt='%1.1d')
    
    TranslateStack(path,TM,bg=bg, extend = extend)
    print("Adaptive template matching done!")  
    plt.show(block=False)
    plt.pause(0.1)
    plt.close()



    
def TranslateStack(movie_path,matrix,bg=150,extend=True):
    '''
    function to apply translation data from a csv
    input: tif stack, matrix[index, tx,ty], background color, extend: whether extend the original image
    output: translated tiff stack
    '''
#    print('''
#=============================================================================
#Please choose the tiff stack you want to translate
#              ''')
#    movie_path = get_path()
    movie_pathout = movie_path[:-4]+'_TF.tif'
    #TS to indicate translated stack
    img = tifffile.imread(movie_path)
    #produces a 3d matrix of nframes, height, width
    nframes, height, width =  img.shape
#    dx, dy, csv_path = get_matrix(csv_path)
    temparray = np.zeros((width, height))

    if extend ==True:
        right= int(abs(max(matrix[:,1])))
        left= int(abs(min(matrix[:,1])))
        top = int(abs(min(matrix[:,2])))
        bottom = int(abs(max(matrix[:,2])))
    else:
        right=0
        left=0
        top=0
        bottom= 0
    height = height+top+bottom
    width = width+left+right
    
    #2d matrix flipped order of width and height to work with warpaffine method
    print("\nApplying translation to stack\n")
    with tifffile.TiffWriter(movie_pathout,bigtiff=True) as tif:

        for num in tqdm(range(nframes)):
            temparray = img[num,:,:]
            TM= np.float32([[1,0,matrix[num,1]],[0,1,matrix[num,2]]])
            #Generates the translation matrix from the dx dy data
    
                
            extend_img = cv2.copyMakeBorder(temparray, top, bottom, left, right, cv2.BORDER_CONSTANT, value=bg)
            translated_img = cv2.warpAffine(extend_img,TM,(width,height),borderValue =bg)#fill with gray
            tif.write(translated_img, contiguous=True)

            
def denoise(csvpath=None):
    '''
    function to denoise a tranlsation matrix using median filtering
    input: none
    output: denoised translation matrix saved as a csv
    Example:
        path = get_path()
        template_match(path)
        denoise()
        apply_translation()
    '''
    dx, dy, csvpath = get_matrix(csvpath)
    csvpathout = csvpath[:-4]+'_denoise.csv'
    mediandx = signal.medfilt(dx,5)
    mediandy = signal.medfilt(dy,5)
    #Kernel set to 5
    index = np.zeros((len(dx),2))
    for num in range(len(dx)):
        index[num,0] = mediandx[num]
        index[num,1] = mediandy[num]
    np.savetxt(csvpathout, index, fmt = '%1.1d', delimiter=",")
    return index


def denoiseTM(TM,tw=5):
    '''
    function to denoise a tranlsation matrix using median filtering
    input: Translation matrix[i,dX,dY],filter width tw
    output: denoised translation matrix saved as a csv
    Example:
        path = get_path()
        template_match(path)
        denoise()
        apply_translation()
    '''
    # dx, dy, csvpath = get_matrix(csvpath)
    # csvpathout = csvpath[:-4]+'_denoise.csv'
    dx=TM[:,1]
    dy=TM[:,2]
    mediandx = signal.medfilt(dx,tw)
    mediandy = signal.medfilt(dy,tw)
    #Kernel set to 5
    
    index = np.zeros((len(dx),3))
    index[:,0]=TM[:,0]
    for num in range(len(dx)):
        index[num,1] = mediandx[num]
        index[num,2] = mediandy[num]
    # # np.savetxt(csvpathout, index, fmt = '%1.1d', delimiter=",")
    return index
    
def stackreg(path = None):
    '''
    Applies SIFT translation to the image stack provided and prints the translation matrix to a csvfile
    input: 
    output: Translation matrix saved as a csv and the csv path
    Example:
        path = get_path()
        stackreg(path)
        apply_translation()
    '''
    from pystackreg import StackReg
    if path == None:
        path = get_path()
    csvpathout = path[:-4] + "_SR.csv"
    img = io.imread(path) # 3 dimensions : frames x width x height
    nframes, height, width = img.shape
    sr = StackReg(StackReg.TRANSLATION)
    TM = sr.register_stack(img, reference = 'previous', verbose = True)
    #Previous seems to be the best for our purposes
    #The ImageJ plugin I believe uses previous
    '''
    reference = (string, one of ['previous', 'first', 'mean']) –
    ‘previous’: Aligns each frame (image) to its previous frame in the stack
    ’first:’ Aligns each frame (image) to the first frame in the stack - if n_frames is > 1, then each frame is aligned to the mean of the first n_frames of the stack
    ’mean’: Aligns each frame (image) to the average of all images in the stack
    '''
    TempTM = np.zeros((nframes, 2))
    for num in range(nframes):
        TempTM[num,0] = TM [num, 0, 2]
        TempTM[num,1] = TM [num, 1, 2]   
    np.savetxt(csvpathout, TempTM, fmt = '%1.1d', delimiter=",")
    return csvpathout
    
def get_roi(imgpath):
    '''
    function to allow roi selection with a gui
    input: image path
    output: region of interest (x,y,w,h), template image, and full image
    '''
    im = cv2.imread(imgpath)
    #only reads the first image in the Tiff stack 
    # Select ROI
    roi = cv2.selectROI(im, False, False)
    img = io.imread(imgpath)
    templ = img[0, int(roi[1]):int(roi[1]+roi[3]), int(roi[0]):int(roi[0]+roi[2])]
    cv2.imshow("Image", templ)
    #Displays the cropped image
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    #closes the cropped image
    return roi, templ, img

def template_match(imgpath, w=24):
    '''
    function to produce a translation matrix based off of template matching
    input: image path
    output: csv with dx and dy in separate columns
    example:
        path = get_path()
        template_match(path)
        denoise()
        apply_translation()
    '''
    import insitu_IO as IO
    from skimage.feature import match_template
    csvpathout = imgpath[:-4] + "_TMatch.csv"
    #Gets the image path
    r, templ, img = get_roi(imgpath)
    #w is the number of pixels around the template that the program will search
    tiffstack = img[:, int(r[1])-w:int(r[1]+r[3])+w, int(r[0])-w:int(r[0]+r[2])+w]
    #Tiffstack is the region that will be searched for matching points
    x = np.zeros((tiffstack.shape[0]), dtype = "int")
    y = np.zeros(tiffstack.shape[0], dtype = "int")
    #Generate numpy arrays to hold the x and y coordinates of the matching points
    iterate = tqdm(range(tiffstack.shape[0]))
    for num in iterate:
        result = match_template(tiffstack[num], templ)
        ij = np.unravel_index(np.argmax(result), result.shape)
        x[num], y[num] = ij[::-1]
    dx = x - w
    dy = y - w
    TM = np.zeros((tiffstack.shape[0],2), dtype='int')
    for num in iterate:
        TM[num,0] = dx[num]
        TM[num,1] = dy[num]
    IO.writeCSV(csvpathout,TM)


def ATM_locglobal(path, NOF=50, csvout=True, TMout=True, TMdenoise=True, TMdenoise_window=5, extend=True, bg=150,search_margin_local=1, match_threshold=0.8):
    '''
    Function to produce adaptive template matching on TIFF image stack based on ROI selection with local and global selections
    
    Input:
        path: Image path for grayscale TIFF stack
        NOF: Number of frames for template to grow
        csvout: Whether to write the translation matrix
        TMout: Whether to write the template match found feature
        TMdenoise: Whether to denoise the measured matrix to remove sudden change
        TMdenoise_window: Window size for TMdenoise
        bg: Background color (0-255)
        extend: Whether to extend the image for the translation (avoid crop)
        match_threshold: Threshold for considering the new ROI position close to the previous one
    
    Output:
        Translated image stack
        CSV file of translation matrix
        Captured ROI location

    Written by Meng Li (mona.mengli@gmail.com) 2024.07.06
    '''

    ###############################################################
    # 1. Initial settings 


    import sys

    try:
        stack = tifffile.imread(path)
    except OSError as e:
        print(f"Error reading the TIFF file: {e}")
        sys.exit(1)

    # stack = tifffile.imread(path)
    nFrames, h, w = stack.shape
    
    TM = np.zeros((nFrames, 3))  # frame, TX, TY, max_val
    pathout3 = path[:-4] + '_TM.csv'
    pathout2 = path[:-4] + '_TM.tif'
    
    ###############################################################
    # 2. Open the first frame and select ROI for template matching
    img = stack[0]
    
    # Select ROI for template matching, returns x1,y1,w,h
    r = cv2.selectROI("Select ROI, press ENTER to confirm", img, fromCenter=False, showCrosshair=False)
    x1,y1,template_w,template_h=int(r[0]),int(r[1]),int(r[2]),int(r[3])
    x2=x1+template_w
    y2=y1+template_h

    cv2.destroyWindow("Select ROI, press ENTER to confirm")
    
    # Crop image to get the template
    # template = img[int(r[1]):int(r[1] + r[3]), int(r[0]):int(r[0] + r[2])]
    template=img[y1:y2,x1:x2]
    # template_h, template_w = template.shape
    
    # Display cropped image
    cv2.imshow("Template, press ENTER to confirm", template)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    cv2.waitKey(1)
    
    ###############################################################
    # 3. Apply template matching
    method = cv2.TM_CCOEFF_NORMED
    
    print("------------------------------------------")
    print("Calculating drift~")  
    
    last_top_left = (x1,y1)
    
    last_max_val = None
    

    
    with tifffile.TiffWriter(pathout2, bigtiff=True) as tif:
        for i in tqdm.tqdm(range(nFrames)):
            fr = stack[i]
            
            search_area_global=fr
 
            # Apply global template matching
            res_global = cv2.matchTemplate(search_area_global, template, method)
            min_val_global, max_val_global, min_loc_global, max_loc_global = cv2.minMaxLoc(res_global)
            
            global_top_left = max_loc_global
            

            # Define the local search area around the last known ROI position
            # search_margin_local = 1  # Local search margin
            x1_local=int(x1-search_margin_local)
            y1_local=int(y1-search_margin_local)
            x2_local=int(x1+template_w+search_margin_local)
            y2_local=int(y1+template_h+search_margin_local)
            search_area_local = fr[y1_local:y2_local, x1_local:x2_local]
            # Apply local template matching
            res_local = cv2.matchTemplate(search_area_local, template, method)
            min_val_local, max_val_local, min_loc_local, max_loc_local = cv2.minMaxLoc(res_local)
            
            local_top_left = (x1_local + max_loc_local[0], y1_local + max_loc_local[1])

   

            # # start_x_local = max(last_top_left[0] - search_margin_local, 0)
            # # start_y_local = max(last_top_left[1] - search_margin_local, 0)
            # # end_x_local=last_top_left[0]+template_w+search_margin_local
            # # end_y_local = last_top_left[1] +template_h + search_margin_local
            # # # end_x_local = min(last_top_left[0]+template_w + search_margin_local, w)
            # # # end_y_local = min(last_top_left[1] +template_h + search_margin_local, h])
            
            # # search_area_local = fr[start_y_local:end_y_local, start_x_local:end_x_local]
            
            # # Apply local template matching
            # res_local = cv2.matchTemplate(search_area_local, template, method)
            # min_val_local, max_val_local, min_loc_local, max_loc_local = cv2.minMaxLoc(res_local)
            
            # local_top_left = (start_x_local + max_loc_local[0], start_y_local + max_loc_local[1])
            
            # Use the local match if the match value is within the threshold of the previous match
            if last_max_val is not None and max_val_local >= max_val_global * match_threshold:
                top_left = (local_top_left[0] , local_top_left[1])
                max_val = max_val_local
            else:
                top_left = (global_top_left[0], global_top_left[1])
                max_val = max_val_global
            
            TM[i, 0] = i
            TM[i, 1] = top_left[0]
            TM[i, 2] = top_left[1]
            
            # Update the last known position and match value
            last_top_left = top_left
            last_max_val = max_val
            
            # Draw rectangle for visualization
            bottom_right = (top_left[0] + template_w, top_left[1] + template_h)
            cv2.rectangle(fr, top_left, bottom_right, 255, 2)
            
            # Periodically update the template
            if i % NOF == NOF - 1:
                template = fr[top_left[1]:bottom_right[1], top_left[0]:bottom_right[0]]
                template_h, template_w = template.shape
                
            if TMout:
                tif.write(fr, contiguous=True)
    
    TM[:, 1] = TM[0, 1] - TM[:, 1]
    TM[:, 2] = TM[0, 2] - TM[:, 2]
    
    ###############################################################
    # 4. Apply translation matrix to movie
    if TMdenoise:
        TM = denoiseTM(TM, TMdenoise_window)
        Mx = int(np.mean(TM[:, 1]))
        My = int(np.mean(TM[:, 2]))
        TM[:, 1] -= Mx
        TM[:, 2] -= My
    
    print("------------------------------------------")
    print("Saving files~")   
    
    plt.plot(TM[:, 0], TM[:, 1], 'b', TM[:, 0], TM[:, 2], 'r')

    if csvout:
        IO.writeCSV(pathout3, TM, fmt='%1.1d')
    
    TranslateStack(path, TM, bg=bg, extend=extend)
    
    print("Adaptive template matching done!")  
    plt.show()





"""
in-situ TEM Toolbox - Diffraction pattern analysis

Assembles of functions related to process FFT and diffraction patterns

Example:
    
    import insitu_Diff as Diff  
    Diff.f2tif(path,1)

Created on Aug 8 2024
@author: Meng Li
"""
import numpy as np
import cv2
# from skimage.feature import blob_dog
import matplotlib.pyplot as plt
# from scipy import ndimage
import math

# import pandas as pd

def next_power_of_2(x):
    return 1 if x == 0 else 2**(x - 1).bit_length()

def pad_to_nearest_power_of_2(image):
    import cv2
    # Get original dimensions
    h, w = image.shape
    size=max(h,w)

    # Calculate the nearest power of 2 dimensions
    new_size = next_power_of_2(size)

    # Calculate padding amounts
    top_pad = (new_size - h) // 2
    bottom_pad = new_size - h - top_pad
    left_pad = (new_size - w) // 2
    right_pad = new_size - w - left_pad

    # Apply padding
    padded_image = cv2.copyMakeBorder(
        image,
        top=top_pad, bottom=bottom_pad,
        left=left_pad, right=right_pad,
        borderType=cv2.BORDER_CONSTANT,
        value=0  # Black color
    )

    return padded_image

def getFFT(im):
    """
    Function to get FFT from grayscale image, auto pad the image to keep the image size  2^n
    Input: grayscale image (np array)
    Output: grayscale FFT with 2^n size, 32 bit
    """

    pad_im=pad_to_nearest_power_of_2(im)


    fft_image = np.fft.fft2(pad_im)
    fft_shifted = np.fft.fftshift(fft_image)
    fft = 20 * np.log(np.abs(fft_shifted))
    return fft

def FFTstack(path):
    """
    Function to get FFT of the stack
    input: path of the image stack
    output: FFT stack with 2^n size 8 bit images
    """
    import tifffile
    import tqdm
    import cv2

    stack = tifffile.imread(path)
    nFrames,h,w = stack.shape #make sure the order of h and w!
    pathout=path[:-4]+'_FFT.tif'

    print("------------------------------------------")
    print('Calculating FFT of each frame')
    with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
        for i in tqdm.tqdm(range(nFrames)):
            fr=stack[i]
            fft=getFFT(fr)
            #convert the 32 bit image to 8 bit
            normalized_image = cv2.normalize(fft, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX)
            # Convert the normalized image to 8-bit
            fft_8bit = np.uint8(normalized_image)
            tif.write(fft_8bit, contiguous=True)

def FFT2Polar(fft):
    """
    Function to convert FFT to polar image, y= 180˚, x= Rho
    Input: grayscale fft image (np array)
    """
    import cv2
    #function to convert fft image to polar coordinates from the center from 0-180 deg
    center = (fft.shape[1] // 2, fft.shape[0] // 2)
    polar_fft = cv2.linearPolar(fft, center, center[0], cv2.WARP_FILL_OUTLIERS)
    # polar_fft= polar_fft[:,:center[1]]
    #resize the image to realistic size
    resized_polar_fft = cv2.resize(polar_fft, (center[0], 360))
        # plt.imshow(resized_polar_fft)
    
    p_fft1=resized_polar_fft[:180,:]
    p_fft2=resized_polar_fft[180:,:]
    p_fft=p_fft1+p_fft2
    return p_fft

def FFTidentify(fft,pathout,scale=1,sigmaMax=2.5,sigmaMin=2.4,thrs=0.96,sigma_ratio=10):
    """
    Function to identify DPs from FFT 
    input: 
        fft: image
        scale: scale of the original image : 1 px  = XXX nm
        parameters for blobs detection: sigmaMax, sigmaMin, thrs, sigma_ratio
        pathout: path for output image (.tif) and csv

    output:
        idnetified image
        indexed csv



    """
    import math
    import matplotlib.pyplot as plt
    from scipy import ndimage
    from skimage.feature import blob_dog, blob_log, blob_doh
    from math import sqrt
    from insituTEM import insitu_IO as IO
    
    h,w = fft.shape
    x_c=int(w/2)
    y_c=int(h/2)
    r=int(x_c/2)   
    if w>320:
        fft_crop=fft[x_c-r:x_c+r,y_c-r:y_c+r]
        image= ndimage.maximum_filter(fft_crop, size=2, mode='wrap')
    else:
        fft_crop = fft
        image= ndimage.maximum_filter(fft_crop, size=3, mode='wrap')




    blobs_dog = blob_dog(image, max_sigma=sigmaMax,min_sigma=sigmaMin, threshold=thrs,sigma_ratio=sigma_ratio,exclude_border=True)
    blobs_dog[:, 2] = blobs_dog[:, 2] * sqrt(2)

    title = 'DP detection'

    fig, axes = plt.subplots(1, 2, figsize=(10, 5), sharex=True, sharey=True)
    ax = axes.ravel()

    ax[0].set_title('FFT_')
    ax[0].imshow(fft_crop)
    # ax[0].set_axis_off()
    ax[1].set_title(title)
    ax[1].imshow(image)


    plt.tight_layout()
    
   
    data=np.zeros((len(blobs_dog),6))
    data[:,0]=blobs_dog[:,1]
    data[:,1]=blobs_dog[:,0]

    for j in range(len(blobs_dog)):
        y=data[j,1]-r
        x=data[j,0]-r

        data[j,2]=math.sqrt(x**2+y**2)
        # if data[j,2] < 10:
        #     data[j,2]=0
        #     theta = 0
        # else:
        #     theta=math.atan(y/x)
        theta=math.atan(y/x)
        data[j,3]=math.degrees(theta)

    data[:,4]=data[:,2]*scale
    data[:,5]=1/data[:,4]

    plt.scatter(data[:,0],data[:,1],marker='+',color='red')
    for i, (x_i, y_i) in enumerate(zip(data[:,0],data[:,1])):
        plt.annotate(str(i+1), (x_i, y_i), textcoords="offset points", xytext=(0,10), ha='center',color='white')
    pathoutcsv=pathout[:-4]+'.csv'
    IO.writeCSV(pathoutcsv,data,fmt='%1.3f')
    plt.savefig(pathout)

# def getFFT(im):
#     """
#     Function to calculate FFT of the image
#     input: im- np array of image data
#     output: fft image
#     """
#     # Get the original dimensions of the image
#     h, w = im.shape

#     size = min(h,w) #in case h w don't match
#     im_crop=im[:size,:size] #crop image 
#     fft_image = np.fft.fft2(im_crop)
#     fft_shifted = np.fft.fftshift(fft_image)
#     fft = 20 * np.log(np.abs(fft_shifted) + 1e-10)  # Add small constant to avoid log(0)
    
#     return fft

def spotFFT(fft, sigmaMax=2.5, sigmaMin=2.45, sigma_ratio=20, thrs=0.98, R_tol=2, fftfilter=3, tolerance=5, display=True):
    """
    Function to get diffraction spot positions in FFT
    Input: fft- FFT image
            sigmaMax, sigmaMin, sigma_ratio, thrs - Parameters for blob detection
            R_tol - tolerance of center spot
            fftfilter- filtersize to denoise FFT to make diffraction spots more visible
            tolerance - tolerance for symmetry matching
    
    Output: 
        Data of all detected spots:    X(px)   Y(px) 
        Image labeling the spots
     """
    from skimage.feature import blob_dog
    import matplotlib.pyplot as plt
    from scipy import ndimage
    import math
  
    import numpy as np
    import pandas as pd

    h, w = fft.shape
    size = max(h, w)
    scale_fft = 1 / (size)
    
    x_c, y_c = int(w / 2), int(h / 2)
    image = ndimage.maximum_filter(fft, size=fftfilter, mode='wrap')
    blobs_dog = blob_dog(image, max_sigma=sigmaMax, min_sigma=sigmaMin, threshold=thrs, sigma_ratio=sigma_ratio, exclude_border=True)
    blobs_dog[:, 2] *= math.sqrt(2)
    

    
    data = []
    blob_set = set()
    
    for j in range(len(blobs_dog)):
        y, x = blobs_dog[j, 1] - y_c, blobs_dog[j, 0] - x_c
        distance = math.sqrt(x**2 + y**2)
        if distance < R_tol:
            continue
        
        data.append([blobs_dog[j, 1], blobs_dog[j, 0]])
        blob_set.add((blobs_dog[j, 1], blobs_dog[j, 0]))
    
    # Ensure symmetry
    for x, y in data:
        sym_x = 2 * x_c - x
        sym_y = 2 * y_c - y
        
        if not any(math.isclose(sym_x, bx, abs_tol=tolerance) and math.isclose(sym_y, by, abs_tol=tolerance) for bx, by in blob_set):
            data.append([sym_x, sym_y])
            blob_set.add((sym_x, sym_y))
    
    data = np.array(data)

    if display == True:
        fig, axes = plt.subplots(1, 2, figsize=(10, 5), sharex=True, sharey=True)
        ax = axes.ravel()
        ax[0].set_title('FFT')
        ax[0].imshow(fft)
        ax[1].set_title('DP detection')
        ax[1].imshow(image)
        plt.tight_layout()
        plt.scatter(data[:, 0], data[:, 1], marker='+', color='red')
        for i, (x_i, y_i) in enumerate(zip(data[:, 0], data[:, 1])):
            plt.annotate(str(i+1), (x_i, y_i), textcoords="offset points", xytext=(0,10), ha='center', color='white')
        
    return data

def DP_mask(fft,circle_radius=10, center_rad=1,sigmaMax=2.5, sigmaMin=2.45, sigma_ratio=20, thrs=0.98, R_tol=2, fftfilter=3, tolerance=5):
    """
    Create a mask image with circles drawn at detected diffraction points.
    
    Parameters:
        image_shape (tuple): Shape of the image (height, width).
        points (np.ndarray): Array of detected points [[x1, y1], [x2, y2], ...].
        circle_radius (int): Radius of the circles to draw.
        center_rad: radius of center circle
        sigma...: parameters for detect DP
    
    Returns:
        np.ndarray: Mask image with drawn circles.
    """

    mask = np.zeros(fft.shape, dtype=np.uint8)
    points= spotFFT(fft,sigmaMax,sigmaMin,sigma_ratio,thrs,R_tol,fftfilter,tolerance,display=False)
    for x, y in points:
        cv2.circle(mask, (int(x), int(y)), circle_radius, 255, thickness=-1)

    if center_rad != 0:
        center=int(fft.shape[0]/2)
        cv2.circle(mask, (center, center), center_rad, 255, thickness=-1)
    return mask

def maskedIFFT(image,mask,mix_ratio=0.3):
    """
    generate masked image from FFT
    Input: original image; mask; mix_ratio for IFFT filtering
    Output: FFT_image,  IFFT image, IFFT filterd image
    Example: fft_image, ifft_image,filtered_image=maskedIFFT(image,mask,mix_ratio)
    """
    h,w=image.shape

    if h !=w:
        d=min(h,w)
        image=image[0:d,0:d]

    # Perform FFT on the image
    fft = cv2.dft(np.float32(image), flags=cv2.DFT_COMPLEX_OUTPUT)
    fft_shifted = np.fft.fftshift(fft)
    fft_image=np.log(1 + cv2.magnitude(fft_shifted[:, :, 0], fft_shifted[:, :, 1])) #using fft_image to apply mask results in empty image
    
    # Ensure the mask image has the same size as the FFT
    mask_image = cv2.resize(mask,(fft_shifted.shape[1], fft_shifted.shape[0]))

    # Apply the mask to the FFT
    masked_fft_shifted = cv2.bitwise_and(fft_shifted, fft_shifted, mask=mask_image)

    # Perform IFFT on the masked FFT
    masked_fft = np.fft.ifftshift(masked_fft_shifted)
    masked_image = cv2.idft(masked_fft)
    masked_image = cv2.magnitude(masked_image[:, :, 0], masked_image[:, :, 1])
    
    #convert to 8bit images
    fft_image=cv2.normalize(fft_image, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)
    ifft_image=cv2.normalize(masked_image, None, 0, 128, cv2.NORM_MINMAX, cv2.CV_8U) #cv2.normalize(src, dst, alpha, beta, norm_type, dtype)
    
    # # Normalize the pixel values to the range of 0-255
    ifft_image_8 = cv2.normalize(ifft_image, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)
    

    # Add IFFT to original image with mix_ratio
    masked_image_normalized = cv2.normalize(ifft_image, None, 0, int(255*mix_ratio), cv2.NORM_MINMAX, cv2.CV_8U)
    image_normalized = cv2.normalize(image, None, 0, int(255*(1-mix_ratio*0.5)), cv2.NORM_MINMAX, cv2.CV_8U)
    
    filtered_image=cv2.add(image_normalized,masked_image_normalized)


    return fft_image, ifft_image, filtered_image


def measureFFT(fft,path,scale,j=0,sigmaMax=2.5,sigmaMin=2.4,thrs=0.97,R_tol=2,fftfilter=2,savefig=True,displayfig=True):
    """
    Function to measure diffraction spots in FFT
    Input: fft- FFT image
            path: path to save data
            sigmaMax, sigmaMin, thrs - Parameters for blob detection
            R_tol - tolerance of center spot
            
    Output: 
        Data of all detected spots: Point   X(px)   Y(px)  distance(px)  angle(˚)  1/d (1/nm)  d (nm)
        Image labeling the spots

    """
    from skimage.feature import blob_dog
    import matplotlib.pyplot as plt
    from scipy import ndimage
    import math
  
    import numpy as np
    import pandas as pd

    h,w = fft.shape
    size=max(h,w)
    scale_fft=1/(scale*size)
    
    x_c=int(w/2)
    y_c=int(h/2)
    center=x_c/2 #center of cropped_fft
    
    fft_crop=fft[x_c-int(center):x_c+int(center),y_c-int(center):y_c+int(center)]
    image= ndimage.maximum_filter(fft_crop, size=fftfilter, mode='wrap')
    blobs_dog = blob_dog(image, max_sigma=sigmaMax,min_sigma=sigmaMin, threshold=thrs,sigma_ratio=10,exclude_border=True)
    blobs_dog[:, 2] = blobs_dog[:, 2] * math.sqrt(2)
    
    
    title = 'DP detection'
    title1= 'FFT_'+str(j)
    fig, axes = plt.subplots(1, 2, figsize=(10, 5), sharex=True, sharey=True)
    ax = axes.ravel()
    
    ax[0].set_title(title1)
    ax[0].imshow(fft_crop)
    ax[1].set_title(title)
    ax[1].imshow(image)
    
    plt.tight_layout()
    
    pathout=path[:-4]+'_'+str(j)+'_FFT.csv'
    pathout_img=path[:-4]+'_'+str(j)+'_DP.tif'
    
    
    data=np.zeros((len(blobs_dog)-1,8))
    
    
    idx=0
    for j in range(len(blobs_dog)):
    
    
        y=blobs_dog[j,1]-center
        x=blobs_dog[j,0]-center
    
        
        distance=math.sqrt(x**2+y**2)#px distance to center
        if distance<R_tol: #skip center point
            continue

        theta=math.atan2(x,y)#atan gives +-90˚ values, atan2 gives 180
        angle=0-math.degrees(theta)
        data[idx,:7] = [j , blobs_dog[j, 1], blobs_dog[j, 0], distance, angle, distance * scale_fft, 1 / (distance * scale_fft)]
        idx += 1
    
    
    plt.scatter(data[:,1],data[:,2],marker='+',color='red')
    for i, (x_i, y_i) in enumerate(zip(data[:,1],data[:,2])):
        plt.annotate(str(i+1), (x_i, y_i), textcoords="offset points", xytext=(0,10), ha='center',color='white')
    
    
    data[:,7]=data[:,4]-data[0,4]
    
    columns= ["Point","X(px)","Y(px)","distance(px)","angle to X(˚)","1/d (1/nm)","d (nm)","angle to P1(˚)"]
    # Set pandas global display format for floats
    pd.options.display.float_format = '{:.3f}'.format
    # Create a DataFrame
    df = pd.DataFrame(data, columns=columns)
    df.to_csv(pathout,index=False)
    if savefig:
        plt.savefig(pathout_img,bbox_inches='tight', pad_inches=0.1,dpi=150)
    if displayfig:
        plt.show()

    return df




def FFT2Polar(fft):
    """
    Function to convert fft image to polar image
    input: fft pimage 
    output: polar FFt image
    """
    import cv2
    #function to convert fft image to polar coordinates from the center from 0-180 deg
    center = (fft.shape[1] // 2, fft.shape[0] // 2)
    polar_fft = cv2.linearPolar(fft, center, center[0], cv2.WARP_FILL_OUTLIERS)
    # polar_fft= polar_fft[:,:center[1]]
    #resize the image to realistic size
    resized_polar_fft = cv2.resize(polar_fft, (center[0], 360))
        # plt.imshow(resized_polar_fft)
    
    p_fft1=resized_polar_fft[:180,:]
    p_fft2=resized_polar_fft[180:,:]
    p_fft=p_fft1+p_fft2
    return p_fft

def maskedFFT(image,mask):
    """
    generate masked image from FFT
    Input: original image; mask
    Output: FFT_image, filtered IFFT image
    """
    h,w=image.shape

    if h !=w:
        d=min(h,w)
        image=image[0:d,0:d]

    # Perform FFT on the image
    fft = cv2.dft(np.float32(image), flags=cv2.DFT_COMPLEX_OUTPUT)
    fft_shifted = np.fft.fftshift(fft)
    fft_image=np.log(1 + cv2.magnitude(fft_shifted[:, :, 0], fft_shifted[:, :, 1]))
    

    # Ensure the mask image has the same size as the FFT
    mask_image = cv2.resize(mask,(fft_shifted.shape[1], fft_shifted.shape[0]))

    # Apply the mask to the FFT
    masked_fft_shifted = cv2.bitwise_and(fft_shifted, fft_shifted, mask=mask_image)


    # Perform inverse FFT on the masked FFT
    masked_fft = np.fft.ifftshift(masked_fft_shifted)
    masked_image = cv2.idft(masked_fft)
    masked_image = cv2.magnitude(masked_image[:, :, 0], masked_image[:, :, 1])
    
    #convert to 8bit images
    fft_image=cv2.normalize(fft_image, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)
    masked_image=cv2.normalize(masked_image, None, 0, 128, cv2.NORM_MINMAX, cv2.CV_8U) #cv2.normalize(src, dst, alpha, beta, norm_type, dtype)
    


    return fft_image, masked_image


# def FFTstack(stackpath,ndfilersize=3,fps=15):
#     import tqdm
#     from scipy import ndimage
#     import tifffile
#     from insituTEM import insitu_IO as IO
#     """
#     Apply FFT to the entire stack. 
#     """
#     pathout=stackpath[:-4]+'_FFT.tif'
#     stack =tifffile.imread(stackpath)
    
#     nFrames, h,w = stack.shape
#     print("------------------------------------------")
#     print("Getting FFT to the stack:")   
#     with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
#         for i in tqdm.tqdm(range(nFrames)):
#             fr=stack[i]
#             fft=Diff.getFFT(fr)
#             if ndfilersize!=0:
#                 fft = ndimage.maximum_filter(fft, size=ndfilersize, mode='wrap')
#             fft_image=cv2.normalize(fft, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)
#             tif.write(fft_image, contiguous=True)
#     IO.tiff2avi(pathout,fps=15)

#     print("Conversion done!")


"""
in-situ TEM Toolbox - TEM raw data conversion

Assembles of functions related to read and process raw dm4, dm3 or FEI TIA or Ceta datas.

Example:
    
    import insitu_EMRaw as EM  
    IO.f2tif(path,1)

Created on Aug 8 2024
@author: Meng Li
"""
import numpy as np
import tqdm
import ncempy
import tifffile
import os

def img_2_8bit(img):
    """
    Convert image to Uint8bit with custom min and max intensity: vmin to 0, vmax=255

    """
    vmax = int(np.median(img)*2)
    vmin = int(np.min(img))
    if (vmax-vmin)==0:
        scale =1 
    else:
        scale = 255/(vmax-vmin)
    imgOut=(img*scale).astype(np.uint8)
    return imgOut

def getScaleBar(barLenPix,scale):
    """
    Function to convert truescale into str with correct format
    """
    truescale= barLenPix/scale
    
    # Find the nearest power of 10 to the truescale
    nearest_power = 10 ** np.floor(np.log10(truescale))

    # Determine the most appropriate scale bar value
    if truescale >= 5 * nearest_power:
        scalebar = 5 * nearest_power
    elif truescale >= 2 * nearest_power:
        scalebar = 2 * nearest_power
    else:
        scalebar = nearest_power

    # Calculate the scale in pixels
    scalepix = int(scalebar * scale)

    # Determine the appropriate unit for the scale bar text
    if scalebar >= 1000:
        text = f"{int(scalebar/1000)} um"
    else:
        text = f"{int(scalebar)} nm"
    return scalepix, text

def AddScalePix(img,barLenPix =250, scale = 34.5, px=0.02,py=0.96, barthick=5,lw=1,stroke=True,tscale=1,textThick=4):
    """
    Function to add scale to image
    Input:
        img: greyscale image
        scaleLen: desired scale length
        scale: scale of the image: px/nm: 1000kX- 49.5; 700kX - 34.5
        (px,py): position of the scale bar: r(0-1)
        lw: boarder of the text
        stroke=1: add stroke; stroke=0 No stroke
 
    """
    import cv2
    h,w=img.shape

    scalepix, text = getScaleBar(barLenPix, scale)

    # w_scale=int(scale*barLen)
    x1=int(w*px)
    y1=int(h*py)
    x2=x1+scalepix
    y2=y1+barthick


    #find bcakground color and make the reverse as scalebar color
    img_bg=img[y1:y2,x1:x2]
    mean_intensity=np.mean(img_bg)

    if mean_intensity<128:
        color = 255
    else:
        color = 0

    bcolor=255-color

    # text = str(barLen)+' nm'
    font = cv2.FONT_HERSHEY_SIMPLEX#CV_FONT_HERSHEY_SIMPLEX normal size sans-serif font
    if stroke == True:
        cv2.putText(img,text,(x1,y1-8), font, tscale, bcolor,int(textThick+lw), cv2.LINE_AA) #Stroke  
        cv2.rectangle(img,(x1,y1),(x2,y2),bcolor,2)#boarder color

        
    cv2.rectangle(img,(x1,y1),(x2,y2),color,-1)
    
    cv2.putText(img,text,(x1,y1-8), font,tscale, color, textThick, cv2.LINE_AA)


def DM2Tiff(path,addscale = True,scalebar_pix=200, tscale=1.2, textThick=4):
    """
    Function to convert dm4 file to tiff. 
    If the input is .dm4  stacked image, convert the stack to tiff stack.
    If the input is .dm4 single image, convert the image to tiff image.
    If the input is a folder with .dm4 single images, convert the folder into tiff stack.
    """
    import ncempy
    import tifffile



    if path.endswith(('.dm4', '.dm3')):
        
        dm = ncempy.read(path)
        pathout =path[:-4]+'.tif'
     
        if len(dm['data'].shape)==2:
            #single image
            print("convert .dm3/4 file to tiff image")
            #for single image
            frame=dm['data']
            img=img_2_8bit(frame)
    
            if addscale==True: 
            
                scale = 1/(dm['pixelSize'][0]*1e3)
    
                AddScalePix(img,barLenPix =scalebar_pix, scale = scale, px=0.02,py=0.98, barthick=5,lw=1,stroke=True,tscale=tscale,textThick=textThick)
            
            tifffile.imwrite(pathout,img, append=False)
        else:
        #image stacks
            print("convert .dm4 stack to tiff stack")
            imgs=dm['data']
            nFrames,w,h =dm['data'].shape

                 
            with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
                for i in tqdm.tqdm(range(nFrames)):
                    img=imgs[i]
                    # img = cv2.convertScaleAbs(img)
                    # img = (img/ 65535.0 * 255).astype(np.uint8)
                    img=img_2_8bit(img)
    
                    # img =cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)
                    if addscale==True: 
                        scale = 1/(dm['pixelSize'][0]*1e3)
                        AddScalePix(img,barLenPix =scalebar_pix, scale = scale, px=0.02,py=0.98,barthick=5,lw=1,stroke=True,tscale=tscale,textThick=textThick)
                
                    tif.write(img, contiguous=True)

    else: 
        files = [os.path.join(path, f) for f in os.listdir(path) if f.endswith(('.dm4', '.dm3'))]
        #convert each file to tiff stack
        print('Convert folder with dm3/4 to tiff stack')
        nFrames=len(files)
        pathout=path+'.tif'
        with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
            for file_path in tqdm.tqdm(files):
                
                dm = ncempy.read(file_path)
                frame=dm['data']
                img=img_2_8bit(frame)
        
                if addscale==True: 
                
                    scale = 1/(dm['pixelSize'][0]*1e3)
        
                    AddScalePix(img,barLenPix =scalebar_pix, scale = scale, px=0.02,py=0.98, barthick=5,lw=1,stroke=True,tscale=tscale,textThick=textThick)
                
                tif.write(img, contiguous=True)
    print("Convertion done!")


def GMS2folder(root_dir):

    import os
    import shutil
    import re
    import tqdm
    """
    Function to convert GMS saved in situ data into one folder named by frame name

    input: path to root_dir : the folder before Hour_00
    output: dm4 images saved in one folder
    """

    # Define the target directory
    target_dir = root_dir + '_output'
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    
    # Initialize a list to hold file paths and their corresponding timestamps
    images = []
    
    # Regular expression to extract hour, minute, second, and frame number from filenames
    pattern = re.compile(r"Hour_(\d+)_Minute_(\d+)_Second_(\d+)_Frame_(\d+)\.dm4")
    
    print("Converting GMS datasets into one folder...")
    
    # Walk through the directory structure and count frames per second
    fps = 0
    for root, dirs, files in os.walk(root_dir):
        for file in files:
            if file.endswith(".dm4"):  # Check if the file is an image file
                match = pattern.search(file)
                if match:
                    hour, minute, second, frame = match.groups()
                    # Create a timestamp tuple for sorting
                    timestamp = (int(hour), int(minute), int(second), int(frame))
                    # Add the file path and timestamp to the list
                    images.append((os.path.join(root, file), timestamp))
                    # Count frames in the Second_00 folder
                    if int(second) == 0:
                        fps += 1
    
    print(f"Frames per second (fps) detected: {fps}")
    
    # Sort the images by their timestamps
    images.sort(key=lambda x: x[1])
    
    # Copy and rename images to the target directory in the correct order
    for index, (filepath, timestamp) in enumerate(images, start=0):
        new_filename = f"F{index:06d}.dm4"
        new_filepath = os.path.join(target_dir, new_filename)
        shutil.copy(filepath, new_filepath)
    
    print(f"Total number of images: {len(images)}")
    print(f"All images have been saved to: {target_dir}")
    return target_dir,fps

def DMfolder2tiff(folder_path):
    """
    Function to convert entire folder of dm files into tiff
    If the dm file is single image, convert to tiff single image
    If the dm file is stack, convert to tiff stack
    """

    print("Batch convert folder into tiff or movie: ")
    print("----------------------------------------")
    # List all files in the directory
    import os
    files = os.listdir(folder_path)

    # Filter files with .ser extension
    dm_files = [file for file in files if file.endswith(('.dm4', '.dm3'))]

    # Process each .ser file
    for dm_file in dm_files:
        print("convert ", dm_file)
            # Check file size
        
        file_path = os.path.join(folder_path, dm_file)
        file_size = os.path.getsize(file_path)
        if file_size == 0:
            print(f"Skipping empty file: {dm_file}")
            continue
        DM2Tiff(file_path,scalebar_pix=250,tscale=1.2, textThick=4)
    print("----------------------------------------")
    print("Convertion done! ")


# def getScaleBar(barLenPix,scale):
#     """
#     Function to convert truescale into str with correct format
#     """
#     truescale= barLenPix/scale
    
#     # Find the nearest power of 10 to the truescale
#     nearest_power = 10 ** np.floor(np.log10(truescale))

#     # Determine the most appropriate scale bar value
#     if truescale >= 5 * nearest_power:
#         scalebar = 5 * nearest_power
#     elif truescale >= 2 * nearest_power:
#         scalebar = 2 * nearest_power
#     else:
#         scalebar = nearest_power

#     # Calculate the scale in pixels
#     scalepix = int(scalebar * scale)

#     # Determine the appropriate unit for the scale bar text
#     if scalebar >= 1000:
#         text = f"{int(scalebar/1000)} um"
#     else:
#         text = f"{int(scalebar)} nm"
#     return scalepix, text

# def AddScalePix(img,barLenPix =250, scale = 34.5, px=0.02,py=0.96, barthick=5,lw=1,stroke=True,tscale=1,textThick=4):
#     """
#     Function to add scale to image
#     Input:
#         img: greyscale image
#         scaleLen: desired scale length
#         scale: scale of the image: px/nm: 1000kX- 49.5; 700kX - 34.5
#         (px,py): position of the scale bar: r(0-1)
#         lw: boarder of the text
#         stroke=1: add stroke; stroke=0 No stroke
#         tscale: scale of the text
#         textThick: thickness of the text in pxs
 
#     """
#     import cv2
#     h,w=img.shape

#     scalepix, text = getScaleBar(barLenPix, scale)

#     # w_scale=int(scale*barLen)
#     x1=int(w*px)
#     y1=int(h*py)
#     x2=x1+scalepix
#     y2=y1+barthick


#     #find bcakground color and make the reverse as scalebar color
#     img_bg=img[y1:y2,x1:x2]
#     mean_intensity=np.mean(img_bg)

#     if mean_intensity<128:
#         color = 255
#     else:
#         color = 0

    
#     bcolor=255-color


    
#     # text = str(barLen)+' nm'
#     font = cv2.FONT_HERSHEY_SIMPLEX#CV_FONT_HERSHEY_SIMPLEX normal size sans-serif font
#     if stroke == True:
#         cv2.putText(img,text,(x1,y1-8), font, tscale, bcolor,int(textThick+lw), cv2.LINE_AA) #Stroke  
#         cv2.rectangle(img,(x1,y1),(x2,y2),bcolor,2)#boarder color

        
#     cv2.rectangle(img,(x1,y1),(x2,y2),color,-1)
    
#     cv2.putText(img,text,(x1,y1-8), font,tscale, color, textThick, cv2.LINE_AA)
# #Save image to tiff or tiff stack

def convertser(path,addscale = True,scalebar_pct=20,fps=10):
    """
    Convert ser file to image, stack to tiff stack and movie
    input: 
        path: .ser file path
        addscale: whether scale bar is added
        scalebar_pix: roughly how many pixels the scale bar takes
        scalebar_pct: roughly how much percnet the scale bar takes on the image
        fps: for .ser stack, the fps of the original data
        tscale: scale of the text for the scale bar
        textThick: thickness of the text for the scale bar
    Output:
        single .ser image: Tiff image
        .ser stack: Tiff stack  and avi movie
    """
    import ncempy
    import tifffile
    import tqdm
    import cv2

    import os
    os.environ["IMAGEIO_FFMPEG_EXE"] = "/opt/homebrew/bin/ffmpeg"
    import insituTEM.insitu_IO as IO

    # print("convert .ser file to tiff image")
    ser = ncempy.read(path)
    pathout =path[:-4]+'.tif'
 

    if len(ser['data'].shape)==2:
        #single image
        # print("convert .ser file to tiff image")
        #for single image
        img=ser['data']
        img =cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)
        # img = cv2.convertScaleAbs(img)
        # img = (img/ 65535.0 * 255).astype(np.uint8)

        if addscale==True: 
            w= ser['data'].shape[0]
            scalebar_pix = round(w/100) *scalebar_pct
            tscale=round(w/1024)
            textThick=round(w/1024)*2
            
            scale = 1/(ser['pixelSize'][0]*1e9)
       
            # scalebar= int(scalebar_pix*scale/100)*100
            
            AddScalePix(img,barLenPix =scalebar_pix, scale = scale, px=0.02,py=0.98, barthick=5,lw=1,stroke=True,tscale=tscale,textThick=textThick)
        
        tifffile.imwrite(pathout,img, append=False)

    else:
        #image stacks
        # print("convert .ser file to tiff stack")
        imgs=ser['data']
        nFrames,w,h =ser['data'].shape
        scalebar_pix = round(w/100) *scalebar_pct
        # fps = int(input('Input desired output fps:'))
        pathout2 =path[:-4]+'_'+str(fps)+'.avi'    
        codec  = cv2.VideoWriter_fourcc(*'MJPG')
        video_out = cv2.VideoWriter(pathout2, codec , fps, (w,  h),isColor=False)

        with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
            for i in tqdm.tqdm(range(nFrames)):
                img=imgs[i]
                # img = cv2.convertScaleAbs(img)
                # img = (img/ 65535.0 * 255).astype(np.uint8)
                img =cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U)
                if addscale==True: 
                    scale = 1/(ser['pixelSize'][0]*1e9)
                    AddScalePix(img,barLenPix =scalebar_pix, scale = scale, px=0.02,py=0.98,barthick=5,lw=1,stroke=True,tscale=tscale,textThick=textThick)
            
                tif.write(img, contiguous=True)
                video_out.write(img)




def SerFolder2Tiff(folder_path,fps=15):
    print("Batch convert folder into tiff or movie: ")
    print("----------------------------------------")
    # List all files in the directory
    import os
    files = os.listdir(folder_path)
    
    # Filter files with .ser extension
    ser_files = [file for file in files if file.endswith('.ser')]
    
    # Process each .ser file
    for ser_file in ser_files:
        print("convert ", ser_file)
            # Check file size
        
        file_path = os.path.join(folder_path, ser_file)
        file_size = os.path.getsize(file_path)
        if file_size == 0:
            print(f"Skipping empty file: {ser_file}")
            continue
            
        # Try to convert the SER file, and skip if an error occurs
        try:
            # convertser(file_path, fps=fps, scalebar_pix=scalebar_pix, tscale=tscale, textThick=textThick)
            convertser(file_path, fps=fps, scalebar_pct=20)
    
        except Exception as e:
            print(f"Error converting {ser_file}: {e}")
            continue  # Skip to the next file
        # convertser(file_path,fps=10,scalebar_pix=scalebar_pix,tscale=tscale, textThick=textThick)
    
    print("----------------------------------------")
    print("Convertion done! ")


"""
in-situ TEM Toolbox - DENS holder data process

Assembles of functions related to process raw data by DENS holders

Example:
    
    from insitu_TEM import insitu_DENS as DENS
    DENS.XXX

Created on Tue May 13 2025
@author: Meng Li mli4@bnl.gov

"""

import pandas as pd

def read_pstrace_csv(filepath):
    # Try several encodings until one works
    for encoding in ['utf-16', 'utf-8', 'ISO-8859-1']:
        try:
            with open(filepath, 'r', encoding=encoding) as f:
                lines = [line.strip() for line in f if line.strip()]
            break
        except UnicodeDecodeError:
            continue
    else:
        raise UnicodeDecodeError("Failed to decode file using common encodings.")

    # Detect measurement type (e.g., CV I vs E Scan 1)
    measurement_type = None
    for line in lines:
        if "vs" in line and not line.lower().startswith("date"):
            measurement_type = line
            break

    # Detect actual column header (like "V,µA" or "s,V")
    header_index = None
    for i, line in enumerate(lines):
        if ',' in line and all(any(c.isalpha() for c in token) for token in line.split(',')):
            header_index = i
            break

    if header_index is None:
        raise ValueError("Could not find a valid data header (e.g., 'V,µA' or 's,V')")

    headers = [h.strip() for h in lines[header_index].split(',')]
    data_lines = lines[header_index + 1:]

    # Read and convert numeric data only
    data = []
    for line in data_lines:
        try:
            row = [float(val) for val in line.split(',')]
            data.append(row)
        except ValueError:
            continue  # skip non-numeric lines

    df = pd.DataFrame(data, columns=headers)
    return measurement_type, df


import matplotlib.pyplot as plt

def plot_multiple_xy(datasets, labels=None, xlabel=None, ylabel=None, title=None):
    """
    Plot multiple datasets (each as a 2-column DataFrame) on one plot.

    Args:
        datasets: list of DataFrames with 2 columns
        labels: list of labels for each line
        xlabel, ylabel: optional axis labels
        title: optional plot title
    """
    fig, ax = plt.subplots(figsize=(6, 4), dpi=300)

    colors = ['blue', 'green', 'red', 'orange', 'purple']

    for i, df in enumerate(datasets):
        x = df.iloc[:, 0]
        y = df.iloc[:, 1]
        label = labels[i] if labels else None
        color = colors[i % len(colors)]
        ax.plot(x, y, label=label, color=color, linewidth=1)

    # X and Y = 0 guide lines
    ax.axhline(0, color='gray', linestyle='-', linewidth=0.5)
    ax.axvline(0, color='gray', linestyle='-', linewidth=0.5)

    # Axis style: boxed
    for side in ['top', 'bottom', 'left', 'right']:
        ax.spines[side].set_linewidth(1)
        ax.spines[side].set_color('black')

    # Axis labels
    ax.set_xlabel(xlabel if xlabel else datasets[0].columns[0], fontsize=12, fontweight='bold', labelpad=10)
    ax.set_ylabel(ylabel if ylabel else datasets[0].columns[1], fontsize=12, fontweight='bold', labelpad=10)

    # Axis limits
        # Extract x/y and lengths
    x_list = [df.iloc[:, 0].values for df in datasets]
    y_list = [df.iloc[:, 1].values for df in datasets]
    all_x = np.concatenate(x_list)
    all_y = np.concatenate(y_list)
    # ax.set_xlim(all_x.min()*1.1, all_x.max()*1.1)
    # ax.set_ylim(all_y.min()*1.2, all_y.max()*1.1)
        # Add 10% margin to x and y limits
    x_margin = (all_x.max() - all_x.min()) * 0.05
    y_margin = (all_y.max() - all_y.min()) * 0.05
    
    ax.set_xlim(all_x.min() - x_margin, all_x.max() + x_margin)
    ax.set_ylim(all_y.min() - y_margin, all_y.max() + y_margin)


    # Ticks and layout
    ax.tick_params(direction='in', top=True, right=True)
    ax.grid(False)

    # Title and legend
    if title:
        ax.set_title(title, fontsize=12,fontweight='bold', pad=10)
    if labels:
        ax.legend()

    plt.tight_layout()
    plt.show()


import matplotlib.pyplot as plt

def plot_xy(df, measurement_type=None):
    """
    Plot single xy curve
    """
    x_col = df.columns[0]
    y_col = df.columns[1]
    x = df[x_col]
    y = df[y_col]

    fig, ax = plt.subplots(figsize=(6,4),dpi=300)  # square aspect ratio

    # Plot data
    ax.plot(x, y, linewidth=1, color='blue')

    # Keep left and bottom axis
    ax.spines['left'].set_visible(True)
    ax.spines['bottom'].set_visible(True)

    # Show top and right spine to make a box
    ax.spines['top'].set_visible(True)
    ax.spines['right'].set_visible(True)

    # Draw all 4 spines in black (like a square)
    for side in ['top', 'bottom', 'left', 'right']:
        ax.spines[side].set_linewidth(1)
        ax.spines[side].set_color('black')

    # Add x=0 and y=0 guide lines
    ax.axhline(0, color='orange', linestyle='-', linewidth=0.5)
    ax.axvline(0, color='orange', linestyle='-', linewidth=0.5)

    # Axis label styling
    ax.set_xlabel('Voltage (V)', fontsize=12, fontweight='bold', labelpad=10)
    ax.set_ylabel('Current (uA)' , fontsize=12, fontweight='bold', labelpad=10)

    # Title
    if measurement_type:
        ax.set_title(measurement_type, fontsize=12, fontweight='bold', pad=10)

    # Tick styling
    ax.tick_params(direction='in', top=True, right=True)

    # Turn off grid
    ax.grid(False)

    plt.tight_layout()
    plt.show()


# Convert the plot to movie

import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np
import cv2
import tifffile
from matplotlib.backends.backend_agg import FigureCanvasAgg
import tqdm

def animate_CVplots(datasets, labels=["Scan 1", "Scan 2", "Scan 3"], xlabel="Voltage (V)", ylabel="Current (µA)", title="CV Scan",
                       fps=60, save_basename="cv_animation"):
    """
    Function to animate the plot
      datasets,             # List of DataFrames, each with 2 columns (x, y) for a scan
    labels=None,          # Optional list of labels for each scan (used in legend)
    xlabel=None,          # Optional label for x-axis (default: first column name)
    ylabel=None,          # Optional label for y-axis (default: second column name)
    title=None,           # Optional plot title
    fps=50,          # fps of the input data. Original speed: 4 fps
    save_basename="cv_animation"  # Prefix for saved output files (AVI + TIFF stack)

    """
     # Setup plot and canvas
    fig, ax = plt.subplots(figsize=(6, 4), dpi=150)
    fig.subplots_adjust(left=0.15, right=0.95, top=0.9, bottom=0.15)
    canvas = FigureCanvasAgg(fig)

    # Extract x/y and lengths
    x_list = [df.iloc[:, 0].values for df in datasets]
    y_list = [df.iloc[:, 1].values for df in datasets]
    lengths = [len(x) for x in x_list]
    cum_lengths = np.cumsum(lengths)
    total_frames = cum_lengths[-1]

    # Axis formatting
    for side in ['top', 'bottom', 'left', 'right']:
        ax.spines[side].set_linewidth(1)
        ax.spines[side].set_color('black')
        
    ax.axhline(0, color='gray', linewidth=0.5)
    ax.axvline(0, color='gray', linewidth=0.5)
    ax.tick_params(direction='in', top=True, right=True)
    ax.set_xlabel(xlabel or datasets[0].columns[0], fontsize=12, fontweight='bold', labelpad=10)
    ax.set_ylabel(ylabel or datasets[0].columns[1], fontsize=12, fontweight='bold', labelpad=10)
    if title:
        ax.set_title(title, fontsize=10)

    # Prepare plot elements
    colors = ['blue', 'green', 'red', 'purple']
    lines, tips = [], []
    for i in range(len(datasets)):
        line, = ax.plot([], [], color=colors[i % len(colors)], label=(labels[i] if labels else None), linewidth=1)
        tip, = ax.plot([], [], color=colors[i % len(colors)], marker='o', markersize=4)
        lines.append(line)
        tips.append(tip)

    # Add live current value in top-left
    value_text = ax.text(0.02, 0.95, '', transform=ax.transAxes,
                         fontsize=10, color='black', ha='left', va='top')


    if labels:
        ax.legend()

    # Axis limits
    all_x = np.concatenate(x_list)
    all_y = np.concatenate(y_list)

    # Add 10% margin to x and y limits
    x_margin = (all_x.max() - all_x.min()) * 0.05
    y_margin = (all_y.max() - all_y.min()) * 0.05
    
    ax.set_xlim(all_x.min() - x_margin, all_x.max() + x_margin)
    ax.set_ylim(all_y.min() - y_margin, all_y.max() + y_margin)
    # ax.set_xlim(all_x.min()*1.1, all_x.max()*1.1)
    # ax.set_ylim(all_y.min()*1.2, all_y.max()*1.1)

    # Video writer setup
    canvas.draw()
    width, height = canvas.get_width_height()
    # fps=int(1000/interval)
    aviout=save_basename+"_"+str(fps)+".avi"
    out = cv2.VideoWriter(aviout, cv2.VideoWriter_fourcc(*'MJPG'), fps, (width, height))
    # tiff_frames = []

    # Setup TIFF writer with BigTIFF support
    tif_writer = tifffile.TiffWriter(f"{save_basename}.tif", bigtiff=True)
    
    # Render frames
    print("Plotting the curves into movies~")
    for frame in tqdm.tqdm(range(total_frames)):
        ax.set_title(title, fontsize=12,fontweight='bold')  # reset title in case of matplotlib bugs

        for i, (x, y, line, tip) in enumerate(zip(x_list, y_list, lines, tips)):
            start = cum_lengths[i - 1] if i > 0 else 0
            end = cum_lengths[i]
            if frame < start:
                # Not started yet
                line.set_data([], [])
                tip.set_data([], [])
            elif frame >= end:
                # Completed
                line.set_data(x, y)
                tip.set_data([], [])
            else:
                # Active scan
                idx = frame - start + 1
                line.set_data(x[:idx], y[:idx])
                tip.set_data([x[idx - 1]], [y[idx - 1]])
                value_text.set_text(
                    f"{labels[i] if labels else f'Scan {i+1}'}\nV = {x[idx - 1]:.3f} V\nI = {y[idx - 1]:.3f} µA"
                )

        # Render and convert to image
        canvas.draw()
        buf = np.frombuffer(canvas.tostring_rgb(), dtype=np.uint8)
        image = buf.reshape((height, width, 3))
        image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        out.write(image_bgr)
        # Save TIFF frame (with compression)
        tif_writer.write(image, photometric='rgb', compression='deflate')

    out.release()
    tif_writer.close()

    # tifffile.imwrite(f"{save_basename}.tif", tiff_frames, photometric='rgb',compression='deflate')
    print("Conversion done!")
    print(f"Saved AVI: {save_basename}.avi")
    print(f"Saved TIFF stack: {save_basename}.tif")

from tkinter import Tk, filedialog
def select_multiple_files(filetypes=(("csv files", "*.csv"), ("All files", "*.*"))):
    Tk().withdraw()  # Hide the root window
    filepaths = filedialog.askopenfilenames(title="Select PSTrace .csv files", filetypes=filetypes)
    return list(filepaths)  # Convert from tuple to list



# -*- coding: utf-8 -*-
"""

in-situ TEM Toolbox - PreProcess
By: Meng Li and Michael
Assembles of functions related to movie preprocess 
Example:
    
    import insitu_Preprocess as PP
    PP.blur_score(image)

Created on Tue Jun 11 10:25:25 2019


"""
import numpy as np
import cv2
from scipy import signal
from PIL import Image 
import matplotlib.pyplot as plt
import tifffile
import tqdm

def estimate_dup(A,B,threshold=1):
    """
    Function to detect if two images are duplicated
    """
    D=np.subtract(A,B)
    score=np.mean(D)#mean works better than sum due to noises
    return bool(score<threshold)


def estimate_blank(image,threshold=50):
    """
    funciton to detect blank frames
    returns 1 if is blank
    """
    score = np.sum(image[10:15,:])
    
    return bool(score < threshold)

def blur_score(image):
    """
    funciton to get blur_score of image
    returns blur score
    """
    if image.ndim == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    
    blur_map = cv2.Laplacian(image, cv2.CV_64F)
    score = np.var(blur_map)
    return score

def median_blur (score_list, i,thrs=150,wid=21):
    """
    funciton to detect if frame i is blur
    input: score_list: 1D array of scores; i: frame number; thrs: tolerance of blur threshold; wid： median filter width
    returns 1 if is blur
    """
    medthreshold=signal.medfilt(score_list,wid)
    #The threshold should be moved because there is no need to recreate the threshold everytime
    score = score_list[i]
    medthres = medthreshold[i]
    return medthres,bool(abs(score-medthres)>=thrs)

    
def detect_blur_fft(image, size=20, thresh=15,vis=False):
    """
    Function to detect if one frame is blurred using FFT
    Example:
    score,is_blur = PP.detect_blur_fft(image, size=20, thresh=15,vis=False)
    Input:
    image: grayscale image, default size: 128*128 px
size: filter size to get the background of FFT
    thresh: threshold value for blur scores, smaller score, more blurry
    Output:
    mean: mean value of the frame
    is_blur: whether the frame is blur
    """
    from insituTEM import insitu_DP as DP
    # grab the dimensions of the image and use the dimensions to derive the center (x, y)-coordinates
    # h, w = image.shape
  

    magnitude=DP.getFFT(image)
    h,w = magnitude.shape
    cX, cY = w //2, h//2
    bar=int(h/100)
    center=bar*10
        #use only the center of the FFT
    crop_h,crop_w=(h-cX)//2, (w-cY)//2
    

    # compute the background of FFT to make the diffraaction spots more obvious
    bg = cv2.blur(magnitude,(size,size))

    enhanced =cv2.subtract(magnitude, bg)
    #remove noise by mean
    # enhanced=enhanced[crop_h:(crop_h+cY),crop_w:(crop_w+cX)]

    #remove white noise from the substracted image
    enhanced[enhanced<20]=0
    
    # Check the data type and convert if necessary
    if enhanced.dtype != np.uint8:
        # Normalize the image to the range [0, 255] if it has a different range
        enhanced = cv2.normalize(enhanced, None, 0, 255, cv2.NORM_MINMAX)
        # Convert the image to uint8
        enhanced = enhanced.astype(np.uint8)
    
    # Apply median filter
    enhanced = cv2.medianBlur(enhanced, 3)

    # enhanced = cv2.medianBlur(enhanced, 3)
    # #remove center spot
    enhanced[(cY - center): (cY + center), cX - center:cX + center] =0
    enhanced[(cY - bar): (cY + bar), :]=0
    enhanced[:, (cX - bar): (cX + bar)]=0
    # enhanced[


    croppedFFT=enhanced[crop_h:(crop_h+cY),crop_w:(crop_w+cX)]

    if vis:
        (fig, ax) = plt.subplots(1, 2,figsize=(10, 5) )
        ax[0].imshow(image, cmap="gray")
        ax[0].set_title("Input")
        # ax[0].set_xticks([])
        # ax[0].set_yticks([])
        # display the magnitude image
        ax[1].imshow(croppedFFT,cmap="viridis")
        ax[1].set_title("Filtered FFT")
        # ax[1].set_xticks([])
        # ax[1].set_yticks([])
        
        # show our plots
        plt.show()
    # mean = np.mean(enhanced)*1000
    mean = np.mean(croppedFFT)*1000
#     print(mean)
    # the image will be considered "blurry" if the mean value of the magnitudes is less than the threshold value
    return (mean,mean<thresh)
        
def tiff_crop(moviepath):
    '''  
    function to crop a tiff stack
    input: path
    output: cropped tiff stack
    '''
    from skimage import io
    from tqdm import tqdm
    import tifffile
    moviepathout = moviepath[:-4] + "_Crop.Tif"
    im = cv2.imread(moviepath)
    #only reads the first image in the Tiff stack
    # Select ROI
    r = cv2.selectROI(im, False, False)
    imCrop = im[int(r[1]):int(r[1]+r[3]), int(r[0]):int(r[0]+r[2])]
    cv2.imshow("Image", imCrop)
    #Displays the cropped image
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    
    #Read the Tiff file properly
    im = io.imread(moviepath)
    nframes, height, width = im.shape
    iterable = range(nframes)
    
    for num in tqdm(iterable): 
        # Crop image
        imCrop = im[num, int(r[1]):int(r[1]+r[3]), int(r[0]):int(r[0]+r[2])]
     
        # Display cropped image
        
        #save cropped image
        if num == 0:                    
            tifffile.imsave(moviepathout, imCrop, append=False)
        else:
            tifffile.imsave(moviepathout, imCrop, append=True)
            
def rotateFr(im,angle=0,expand=True,bgColor=150):
    """
    funciton to rotate frame clockwise by angle degree with filled color
    input: im-greyscale image frame
    returns rotated image
    """
#    from PIL import Image
#    angle = 73
    a=360-angle
    img = Image. fromarray(im)
    fr=img.rotate(a,expand=expand,fillcolor=bgColor)
    imgRt = np.array(fr) 
    return imgRt

def removeDP(img,position):
    """
    funciton to  remove deadpoint in greyscale image using automated color based on surounding pixels

    input: im-greyscale tiff stack
    position:position of DP[x1 y1 w h], or DPs[[x1 y1 w1 h1],[x2 y2,w2,h2],...]
    returns rotated image
    """
    #remove deadpoint in greyscale image using automated color based on surounding pixels
    #position:[x1 y1 w h]
    
    
    x1=position[0]
    y1=position[1]
    w=position[2]
    h=position[3]

    x2=x1+w
    y2=y1+h
    c1=np.mean(img[y1-1,x1:x2])
    c2=np.mean(img[y2+1,x1:x2])
#    c3=np.mean(img[x1-1,y1:y2])
#    c4=np.mean(img[x2+1,y1:y2])
#    color=int((c1+c2+c3+c4)/4)
    color = int((c1+c2)/2)
#    color = 177
    cv2.rectangle(img,(x1,y1),(x2,y2),color,-1)  
    
def removeDPs(img,DP):
    """"
    Function to remove multiple DPs in one image
    DP[[x1 y1 w1 h1],[x2 y2,w2,h2],...]
    """
    nDP=len(DP)
    if nDP==1:
        removeDP(img,DP)
    else:
        for j in range(nDP):
            position=DP[j]
            removeDP(img,position)


def BlurDetection(path,wid=31):
    """
    Function to remove blurry frames based on the blur score
    Input: tif file path
    Output: deblurred tif stack
    """
    from insituTEM import insitu_IO as IO

    indexout=path[:-4]+'_DB-index.csv'
    DBout=path[:-4]+'_DB.csv'
    DBout2=path[:-4]+'_Blur.csv'

    stack = tifffile.imread(path)
    nFrames,h,w = stack.shape #make sure the order of h and w!
    index = []#index of remaining frames
    score = []

    print("------------------------------------------")
    print("Calculating blur scores!")  

    for i in tqdm.tqdm(range(nFrames)):
        fr=stack[i]
        s = blur_score(fr)
        score.append(s)
        index.append(i)
        
    length = len(score)
    Blur=np.zeros((length,3))#[i,blur_score,blur_th]
    Blur[:,0]=index
    Blur[:,1]=score            

    plt.plot(score,'gray')
    plt.show()     

    print("------------------------------------------")
    print("Check blur detection!")  
  
    thrs=int(input( 'Input blur threshold(default: 70): ')) 
    while thrs!=0:
        index2=[]

        for i in tqdm.tqdm(range(len(score))):
            medthres,is_blur=median_blur (score, i,thrs,wid=wid)#wid must be odd
    #        DB[index[i],3]=is_blur
            Blur[i,2]=medthres
            if is_blur == 0:
                index2.append(index[i])
                
        plt.plot(Blur[:,0],Blur[:,1],'gray',Blur[:,0],Blur[:,2]-thrs,'r',Blur[:,0],Blur[:,2]+thrs,'r')
        plt.show()
        thrs=int(input( 'Press 0 if you are satisfied, press corrected threshold if you want to change:  ')) 
    
    IO.writeCSV(DBout2,Blur)  
        
    result= np.zeros((len(index2),2)) 
    for i in range(len(index2)):
        result[i,0]=i
        result[i,1]=index2[i]
        
    IO.writeCSV(indexout,result)

    print("------------------------------------------")
    print("Save blur removed file") 
    pathout2=path[:-4]+'_DBlur.tif'

    with tifffile.TiffWriter(pathout2,bigtiff=True) as tif:
        for j in tqdm.tqdm(range(len(index2))):
            i=int(result[j,1])
            fr=stack[i]

            tif.write(fr, contiguous=True)
    print("------------------------------------------")
    print("Blur removed!") 
    print("==========================================")


def movie_preprocess(path,remove_DP=True,DP=[[0, 0, 1, 1],[0,0,2,2]],Crop=True,cropboarder=[0,0,0,0],Trim=False,start_time_mmss="00:00",end_time_mmss="00:00",grayscale=True,exp_time=0):
    """
    Function to preprocess the raw TEM movie to tiff
    Input:
        path: file path of the movie
        remove DP: if there are constant dead point in the movie to remove, 
            if so, input the DP value. 
        Crop: if crop the movie to remove the boarders
            if so, input the cropboarder: [left, top, right, bottom]
        Trim: if trim the movie length
            if so, input the start and end time in mm:ss
        grayscale: if convert the movie to grayscale
        exp_time: if remove the duplicated frame based on the real exposure time. 
            if no need to remove duplicates, put exp_time = 0
    """
    # print("=========================================")
    print("ETEM movie preprocessor start!")

    import moviepy.editor as mp
    import moviepy.video.fx.all as vfx
    from insituTEM import insitu_IO as IO


    video = mp.VideoFileClip(path)
    if Trim == True:
        # Convert mm:ss to seconds
        start_time_seconds = sum(x * int(t) for x, t in zip([60, 1], start_time_mmss.split(":")))
        end_time_seconds = sum(x * int(t) for x, t in zip([60, 1], end_time_mmss.split(":")))

        video = video.subclip(start_time_seconds, end_time_seconds)# cut the clip between t=0 and 28 secs.

    fps = int(video.fps)
    # fps= 15
    w = int(video.w)
    h = int(video.h)
    nFrames = int(fps*video.duration)

    print("fps=", fps, "   w =",w, "   h=", h, "duration= ",video.duration)
    print("Total number of Frames=", nFrames)



    #Preview first frame
    print("------------------------------------------")
    print("Preview the processed frame: press 0 if you want to continue: ")

    if Crop == True:
        video =  video.fx(vfx.crop, x1=cropboarder[0], y1=cropboarder[1], x2=w-cropboarder[2], y2=h-cropboarder[3])#crop movie snapshot
    
    fr = video.get_frame(mp.cvsecs(1/fps))
    
    if grayscale==True: 
        fr = cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY)

    if remove_DP==True:
        removeDPs(fr,DP)

    IO.showFr(fr)
    prompt = input("Input 0 if you want to continue, other keys to break :")

    if prompt != "0":
        print("Stopped by user. Modify the cropboarder values and restart.")
        return
    else: 
        #continue to apply to the full movie
        #remove duplicate frames with known fps conversion
        if exp_time==0:
            dup=1
        else:
            dup= fps*exp_time

        nFramesOut=int(nFrames/dup)
        print ('nFramesOut= ', nFramesOut)
        print('nFramesIn= ',nFrames)


        pathout=path[:-4]+'_CutDB.tif'

        print("------------------------------------------")
        print("Saving duplicate removed processed frames~")   
        

        with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
            for j in tqdm.tqdm(range(nFramesOut)):
                i=j*dup
                fr = video.get_frame(mp.cvsecs(i/fps))
                if grayscale==True: 
                    fr = cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY)
                if remove_DP==True:
                    removeDPs(fr,DP) 
                tif.write(fr, contiguous=True)

        print("Conversion done!")


# -*- coding: utf-8 -*-
"""
in-situ TEM Toolbox - Input Output

Assembles of functions related to movie input output and format conversion.

Example:
    
    import insitu_IO as IO
    IO.f2tif(path,1)

Created on Tue Jun 11 10:25:25 2019
@author: Meng Li
Update 20200303:
        Add stroke to scalebar and text
"""


import cv2
import tifffile
import tqdm #not necessary just provides a progress bar and timer
import numpy as np
import moviepy.editor as mp

def tiff2mp4(path,fps):
    """
    function to convert any input file to H264 High quality mp4 using openCV
    Inputs: filepath
    Output: file in the same folder named '..._fps.mp4'
    """
    video = tifffile.imread(path)
    nFrames, h,w = video.shape[:3]
    # dur=1/fps    
    pathout =path[:-4]+'_'+str(fps)+'.mp4'    
    # pathout2 =path[:-4]+'_St.tif'
    codec  = cv2.VideoWriter_fourcc(*'H264')
    out = cv2.VideoWriter(pathout, codec , fps, (w,  h))
    print("---------------------------------------------")
    print('Converting Tiff stack to mp4')    
    for i in tqdm.tqdm(range(nFrames)):        
        img=video[i]  
        out.write(img)
    out.release()
    cv2.destroyAllWindows()
    print("==============================================")
    print("MP4 convertion Done!")

def tiff2avi(path,fps,is_color=False):
    #convert tiff stack to movie
    import tifffile
    import cv2
    
    # path = 
    stack = tifffile.imread(path)
    nFrames,h,w = stack.shape[:3]
    
    # # fps = int(input('Input desired output fps:'))
    # # dur=1/fps    
    # pathout =path[:-4]+'_'+str(fps)+'.avi'    
       
    
    # codec  = cv2.VideoWriter_fourcc(*'MJPG')
    
    # video_out = cv2.VideoWriter(pathout, codec , fps, (w,  h))
    # print("---------------------------------------------")
    # print('Converting Tiff stack to avi')    
    # for i in tqdm.tqdm(range(nFrames)):        
    #     img=stack[i] 
    #     video_out.write(img)
        
    # video_out.release()
    # cv2.destroyAllWindows()
    # print("==============================================")
    # print("AVI convertion Done!")
    size=(int(w),int(h))


    pathout =path[:-4]+'_'+str(fps)+'.avi'    

    # # Check if the image stack is grayscale or color
    is_color = len(stack.shape) == 4 and stack.shape[3] == 3

    codec  = cv2.VideoWriter_fourcc(*'MJPG')
    video_out = cv2.VideoWriter(pathout, codec , fps, size,isColor=is_color)
    print("---------------------------------------------")
    print('Converting Tiff stack to avi')    
    for i in tqdm.tqdm(range(nFrames)):        
        img=stack[i]

        # Ensure the image is in 8-bit format
        
        if img.dtype != np.uint8:
            img = cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX)
            img = np.uint8(img)
        # Convert RGB to BGR if the image is in color
        if is_color ==True:
            img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        # else:
        #     img = cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX)
        #     img = np.uint8(img)
        # Convert RGB to BGR if the image is in color
        
            
        video_out.write(img)
        
    video_out.release()
    cv2.destroyAllWindows()
    print("==============================================")
    print("AVI convertion Done!")

def movie2tif(path,is_gray=1,exp_time=0.2):
    """
    function to convert any input movie file to tif stack
    Inputs: filepath of movie, is_gray: 1 for grayscale, 0 for rgb
            exp_time= exposure time of real movie, remove dup frame if fps > exp_time
    
    Output: file in the same folder named '..._cv.tif'
    """   
    import moviepy.editor as mp
    import tifffile
    import tqdm
    print("==============================================")
    print("Convert file to tif stack!")
    pathout = path[:-4]+'_'+str(is_gray)+'.tif'    
    video = mp.VideoFileClip(path)
    fps = int(video.fps)
    w = int(video.w)
    h = int(video.h)
    nFrames = int(fps*video.duration)
    dup= fps*exp_time
    nFramesOut=int(nFrames/dup)
    with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
        for j in tqdm.tqdm(range(nFramesOut)):
            i=int(j*dup)
            fr = video.get_frame(mp.cvsecs(i/fps))
            if is_gray==1:
                fr = cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY)
            tif.write(fr, contiguous=True)


    video.reader.close()
        # for fr in tqdm.tqdm(video.iter_frames()):
        #     if is_gray==1:
        #         fr = cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY) 
        #     tif.write(fr,continuous=True)

    print("==============================================")
    print("TIF convertion Done!")
    video.reader.close()# To fix handel error problem

def folder2tif(path,isGray=True):
    """
    Function to convert folder of single images into tif stack
    Input: path: folder path
    Output: tif stack with the same name as the folder in the parent folder.
    """
    import os
    import tifffile
    img_name=os.listdir(path) 
    nFrames= len(img_name)
    pathout=path+'.tif'


    with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
        for i in tqdm.tqdm(range(nFrames)):
            # img_name = f'{i:03d}.tif'
            compare_path=os.path.join(path,f'{i:03d}.tif')
            overlay =cv2.imread(compare_path)
            if isGray==True:
                overlay = cv2.cvtColor(overlay,cv2.COLOR_BGR2GRAY)
            else:
                overlay = cv2.cvtColor(overlay, cv2.COLOR_BGR2RGB) #tifffile usedRGB, cv2 used BGR
        
            tif.write(overlay, contiguous=True)

def f2gif(path,fps):
    """
    function to convert any input file to gif animation
    Inputs: filepath, output fps, is_gray: 1 for grayscale, 0 for rgb
    
    Output: file in the same folder named '..._cv.mp4'
    """
    
    print("==============================================")
    print("Convert file to GIF!")
    pathout = path[:-4]+'_'+str(fps)+'.gif'
    if path.endswith('.tif'):            
#        import tifffile
        im = tifffile.imread(path)
        nFrames, h,w = im.shape
        dur=1/fps
        clip = []
        for i in range(nFrames):
            fr = cv2.cvtColor(im[i],cv2.COLOR_GRAY2RGB)
            clip.append(mp.ImageClip(fr).set_duration(dur))
        video = mp.concatenate_videoclips(clip, method="compose",ismask=False)#ismask=True to make grayscale

    else:
        video = mp.VideoFileClip(path)
        fpsIn = int(video.fps)
        if fps != fpsIn:
            print("Conflict in fps! \n",              "[0] Use fps of input file;\n",              "[1] Use desired fps w/o speedup;\n",
              "[2] Use desired fps w/ speedup:")
            k = input('Input your selection: ')
            if k == 2:
                sf = fps/fpsIn
                video =video.fx(mp.vfx.speedx, sf)# Not working when sf<1
            elif k == 0:
                fps = fpsIn

        video.write_gif(pathout,fps=fps)
        video.reader.close()# To fix handel error problem
#    if path.endswith('.gif'):
#        clip.write_videofile(pathout,fps=fps,codec='libx264', bitrate='32 M',preset='ultrafast')
    print("==============================================")
    print("MP4 convertion Done!")

def gifmp4converter(path,fpsOut):
    """
    Function to convert between mp4 and gif files
    Inputs: file path; Output fps
    Output: mp4/gif file in the same folder named '..._cv.mp4/gif'
        
    Example: 
        gifmp4converter(path,25)
    """
#    import moviepy.editor as mp
    
    print("=========================================")
    print("GIF-MP4 Converter Started!")

    clip = mp.VideoFileClip(path)
    #Get output fps
    fpsIn = int(clip.fps)
    fps=fpsOut
    if fpsOut != fpsIn:
        print("Conflict in fps! \n",
              "[0] Use fps of input file;\n",
              "[1] Use desired fps w/o speedup;\n",
              "[2] Use desired fps w/ speedup:")
        k = input('Input your selection: ')
        if k == 2:
            sf = fpsOut/fpsIn
            fps = fpsOut                
            clip =clip.fx(mp.vfx.speedx, sf)
        elif k == 0:
            fps = fpsIn
         
# Converting formats
    if path.endswith('.gif'):
        pathout = path[:-4]+'_cv'+'.mp4'
        clip.write_videofile(pathout,fps=fps,codec='libx264', bitrate='32 M',preset='ultrafast')
    elif path.endswith('.mp4'):
        pathout = path[:-4]+'_cv'+'.gif'
        clip.write_gif(pathout,fps=fps)
    clip.reader.close()# To fix handel error problem
    print("=========================================")
    print("GIF-MP4 Converter Done!")
    
def GetInfo(path):
    """
    function to get movie info  from any file form: mp4 , tif, gif
    Inputs: filepath
    Output: nFrames, w, h, fps
    Example: w,h,nFrames,fps = GetInfo(path)

    """   
#    import tifffile
##    import tqdm
#    import moviepy.editor as mp
#    import cv2
    if path.endswith('.tif'):
        video = tifffile.imread(path) 
        nFrames, h,w = video.shape
        fps=int(input("Enter desired fps: "))
    else:
        video = mp.VideoFileClip(path)
        fps= int(video.fps)
        nFrames = int(fps*video.duration)
        w = int(video.w)
        h = int(video.h)
        video.reader.close()# To fix handel error problem
    return w,h,nFrames,fps

def RdFr(path,i,is_gray=1):
    """
    function to read frame from any file form: mp4 , tif, gif
    Inputs: filepath, is_gray: 1 for grayscale, 0 for rgb
    
    Output: frame image fr
    Example: 
IO.fr= RdFr(path,i)
cv2.imshow('Frame'+str(i),fr)
cv2.waitKey(0)
cv2.destroyAllWindows()
    """   
#    import tifffile
##    import tqdm
#    import moviepy.editor as mp
#    import cv2
    if path.endswith('.tif'):
        video = tifffile.imread(path)
        fr = video[i]                    
    else:
        video = mp.VideoFileClip(path)
        fps= int(video.fps)
        fr = video.get_frame(mp.cvsecs(i/fps))
        if is_gray == 1:
            fr= cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY)
        video.reader.close()# To fix handel error problem
    return fr

def ExtractFrame(path,framelist,is_gray=1):
    """
    function to Extract frame from any file format: mp4 , tif, gif
    Inputs: filepath, list of frames to be extracted
    Output: frame images saved in original folder
    Example: 
        import insitu_IO as IO
        IO.ExtractFrame(path,[0,1,10,11],0)

    """ 
    if path.endswith('.tif'):
        video = tifffile.imread(path)
    else:
        video = mp.VideoFileClip(path)
        fps= int(video.fps)
    for i in tqdm.tqdm(framelist):
        if path.endswith('.tif'):
            fr = video[i]                  
        else:
            fr = video.get_frame(mp.cvsecs(i/fps))
            if is_gray == 1:            
                fr= cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY)
#        fr=RdFr(path,int(i),is_gray)          
        pathout = path[:-4]+'_F'+str(i)+'.tif'
        tifffile.imwrite(pathout,fr, append=False)
    if path.endswith('.tif')==0:
        video.reader.close()
        
def writeCSV(pathout,data,fmt='%1.1d'):
#    import csv
    import numpy as np
#    (length,width)=np.shape(data)
    np.savetxt(pathout, data, fmt=fmt, delimiter=",")
     
def showFr(im):
    from PIL import Image
    img = Image. fromarray(im)
    img.show() 

#functions for scalebar
def readscale(path,dy=5):
    """
    Function to measure scale from movie first frame
    Input: path - movie path; dy: shift from the bottom in px
    scalebarcolr- 0 for black, 255 for white
    """
    import os
    os.environ["IMAGEIO_FFMPEG_EXE"] = "/opt/homebrew/bin/ffmpeg"
    import moviepy.editor as mp
    import cv2
    
    video = mp.VideoFileClip(path)
    fr = video.get_frame(mp.cvsecs(0))
    fr = cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY)

    # Select ROI using cv2.selectROI()
    roi = cv2.selectROI('Select scalebar ROI, quit with SPACE', fr)
    cv2.waitKey(10)
    cv2.destroyWindow('Select scalebar ROI, quit with SPACE')

    # Crop the image based on the selected ROI
    cropped_image = fr[int(roi[1]):int(roi[1] + roi[3]), int(roi[0]):int(roi[0] + roi[2])]

    # # # Display the cropped image
    showFr(cropped_image)

    import numpy as np

    scalebar_color = int(input("Input the scalebar color,  0-black, 255-white: "))
    dy2=dy
    while dy2 != 0: 
        print(cropped_image[-dy,:])
        dy2=int(input("input 0 if the crop is correct, otherwise input the estimated distance from bottom to the scalebar: "))
        if dy2 !=0:
            dy=dy2
            
    scalebar_px =  np.count_nonzero(cropped_image[-dy,:]==scalebar_color)
    print("scalebar_px=", scalebar_px)
    scale_nm = int(input("Input the scalebar in nm: "))
    scale =scalebar_px/scale_nm
    print("==============")
    print("scale=",scale,"px/nm")
    return scale

def getScalePix(scale,barLenPix=250):
    """
    Function to convert truescale into str with correct format
    Input: 
        barLenPix : pixel length of the scalebar region
        scale: scale: px/nm
    Output: 
        text: string of the scalebar text, in the format such as 100, 200, 500 nm .
        scalepix: pixel distance of the realoutput scalebar
    """
    truescale= barLenPix/scale
    
    # Find the nearest power of 10 to the truescale
    nearest_power = 10 ** np.floor(np.log10(truescale))

    # Determine the most appropriate scale bar value
    if truescale >= 5 * nearest_power:
        scalebar = 5 * nearest_power
    elif truescale >= 2 * nearest_power:
        scalebar = 2 * nearest_power
    else:
        scalebar = nearest_power

    # Calculate the scale in pixels
    scalepix = int(scalebar * scale)

    # Determine the appropriate unit for the scale bar text
    if scalebar >= 1000:
        text = f"{int(scalebar/1000)} um"
    else:
        text = f"{int(scalebar)} nm"
    return scalepix, text

def AddScalePix(img,barLenPix =250, scale = 34.5, px=0.02,py=0.96, barthick=5,lw=1,stroke=True,tscale=1,textThick=4):
    """
    Function to add scale to image
    Input:
        img: greyscale image
        scaleLen: desired scale length
        scale: scale of the image: px/nm: 1000kX- 49.5; 700kX - 34.5
        (px,py): position of the scale bar: r(0-1)
        lw: boarder of the text
        stroke=1: add stroke; stroke=0 No stroke
 
    """
    import cv2
    h,w=img.shape

    scalepix, text = getScalePix(barLenPix, scale)

    # w_scale=int(scale*barLen)
    x1=int(w*px)
    y1=int(h*py)
    x2=x1+scalepix
    y2=y1+barthick


    #find bcakground color and make the reverse as scalebar color
    img_bg=img[y1:y2,x1:x2]
    mean_intensity=np.mean(img_bg)

    if mean_intensity<128:
        color = 255
    else:
        color = 0

    
    bcolor=255-color

    font = cv2.FONT_HERSHEY_SIMPLEX#CV_FONT_HERSHEY_SIMPLEX normal size sans-serif font
    if stroke == True:
        cv2.putText(img,text,(x1,y1-8), font, tscale, bcolor,int(textThick+lw), cv2.LINE_AA) #Stroke  
        cv2.rectangle(img,(x1,y1),(x2,y2),bcolor,2)#boarder color

        
    cv2.rectangle(img,(x1,y1),(x2,y2),color,-1)
    
    cv2.putText(img,text,(x1,y1-8), font,tscale, color, textThick, cv2.LINE_AA)

def AddScale(img,barLen =2, scale = 34.5, px=0.02,py=0.96, color = 255,fontScale=1.5, barthick=5,textthick=2,stroke=1):
    """
    Function to add scale to image
    Input:
        img: greyscale image
        barLen: desired scale length
        scale: scale of the image: px/nm: 1000kX- 49.5; 700kX - 34.5
        (px,py): position of the scale bar: r(0-1)
        color: color of the scalebar
        stroke=1: add stroke with thickness,  stroke=0 No stroke
  
        fontScale: scale of the text, default is 1.
        barthick: scalebar thickness
        textthick: thickness of the text without stroke
 
    """
    h,w=img.shape
    w_scale=int(scale*barLen)
    x1=int(w*px)
    y1=int(h*py)
    x2=x1+w_scale
    y2=y1+barthick

    bcolor=255-color
    
    text = str(barLen)+' nm'
    font = cv2.FONT_HERSHEY_SIMPLEX#CV_FONT_HERSHEY_SIMPLEX normal size sans-serif font
    shifty=int(8*fontScale)
    if stroke != 0:
        strokethick = int(stroke + textthick)
        rectangle_thickness = max(1, strokethick // 2)  # halve it for visual match
        cv2.putText(img, text, (x1, y1 - shifty), font, fontScale, bcolor, strokethick, lineType=cv2.LINE_AA)  # Stroke  
        cv2.rectangle(img, (x1, y1), (x2, y2), bcolor, rectangle_thickness)  # Border color

    cv2.rectangle(img, (x1, y1), (x2, y2), color, -1)
    cv2.putText(img, text, (x1, y1 - shifty), font, fontScale, color, textthick, cv2.LINE_AA)

def AddText(img,text, px=0.98,py=0.96, color = 255,fontScale=0.9,thickness=2,stroke=1):
    """
    Function to add text str to image
    Input:
        img: greyscale image
        text: textstr to display
        (px,py): position of the text: right,middle corner
        fontScale: scale of the text
        thickness: line thickness of the text
  
        stroke=1: add stroke width of 1; stroke=0 No stroke
 
    """
    h,w=img.shape
 
    font = cv2.FONT_HERSHEY_SIMPLEX#CV_FONT_HERSHEY_SIMPLEX normal size sans-serif font
    D,sd=cv2.getTextSize(text, font, fontScale,thickness)
    wt=int(D[0])
    bcolor=255-color #color of stroke
    # bw =lw+1 #stroke thickness
    strokethick=thickness+stroke
 
    x=int(w*px-wt)
    y=int(h*py)
    if stroke !=0:
        cv2.putText(img,text,(x,y), font, fontScale, bcolor, strokethick, cv2.LINE_AA) #Stroke      
    cv2.putText(img,text,(x,y), font, fontScale, color, thickness, cv2.LINE_AA)   

def AddScaleTime2Stack(path,bar,Scale,dt=0.2,fps=15,color=255, time_txt_scale=1.2,textthick=2,stroke=1):
    """
    Function to add scalebar and timestamp to stack
    Input:
        path: path to tiff stack
        bar: scale bar to show in nm
        Scale: scale for the image stack: px/nm
        dt: time interval of each frame
        others: parameters for adjusting scalebar text size
    Output:
        Tiff stack and avi movie of the added scale&time

    Example: 
    AddScaleTime2Stack(path,bar=20,Scale=22.2,dt=0.5,fps=15,color=255, time_txt_scale=1.2)

    """
    stack=tifffile.imread(path)
    nFrames, h,w = stack.shape
    
    pathout =path[:-4]+'_St.tif'    

    print("---------------------------------------------")
    print('Testing positions! ')
    fr=stack[0]
    px=0.96
    py=0.96

    text='0.0 s'
    AddScale(fr,barLen=bar,fontScale=time_txt_scale,scale=Scale,color=color,textthick=textthick,stroke=stroke)
    AddText(fr,text,px=px,py=py,fontScale=time_txt_scale,color=color,thickness=textthick,stroke=stroke)
    showFr(fr)

    prompt = input("Check the position, press 0 if you are satisfied, any other key to exit: ")
    print("---------------------------------------------")

    if prompt == "0":
        print("Adding scalebar and timestamp to full stack")
        with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
        
            for i in tqdm.tqdm(range(nFrames)):        
                img=stack[i]        
                time = i * dt    
                text=str('%02.1f' %time)+' s'  
                AddScale(img,barLen=bar,fontScale=time_txt_scale,scale=Scale,color=color,textthick=textthick,stroke=stroke)
                AddText(img,text,px=px,py=py,fontScale=time_txt_scale,color=color,thickness=textthick,stroke=stroke)        
                
                tif.write(img, contiguous=True)
        
        
        tiff2avi(pathout,fps=fps)
        print("Done adding scalebar and timestamp to tiff stack! ")
    else:
        print("Exit to change settings")

def apply_colormap(image,cmap='viridis'):
    """
    Function to apply false color to grayscale image
    input: 
    Image: grayscale image
    camp: colormap string. such as 'viridis'
    """
    import matplotlib.pyplot as plt
    colormap = plt.get_cmap(cmap)
    colored_image = colormap(image / 255.0)  # Normalize the image and apply colormap
    colored_image = (colored_image[:, :, :3] * 255).astype(np.uint8)  # Convert to 8-bit RGB
    return colored_image

def tiff2avi_falsecolor(path,cmap='viridis',fps=15):
    """
    Function to convert tiff to false color avi movie
    input: path : path of tiff stack
    cmap: colormap of the falsecolor
    """

    import tifffile
    import numpy as np
    import cv2
    import matplotlib.pyplot as plt
    import tqdm

    print("------------------------------------------")
    print("Convert stack to faulse color avi")
    stack = tifffile.imread(path)
    nFrames,h,w = stack.shape #make sure the order of h and w!
    
 
    pathout =path[:-4]+'_c_'+str(fps)+'.avi'    
    
    
    #Check the colormap effect
    fr = stack[0]
    fr_color=apply_colormap(fr,cmap)
    plt.imshow(fr_color)
    
    
    fourcc = cv2.VideoWriter_fourcc(*'MJPG')
    
    video_writer = cv2.VideoWriter(pathout, fourcc, fps, (w, h),isColor=True)
    
    # Process each frame and write to the video
    for i in tqdm.tqdm(range(nFrames)):
        grayscale_image = stack[i]
        colored_image = apply_colormap(grayscale_image,cmap)
        colored_image = cv2.cvtColor(colored_image, cv2.COLOR_RGB2BGR)  # Convert RGB to BGR
    
        video_writer.write(colored_image)
    
    # Release the video writer
    video_writer.release()

def AverageFrame(path,AveN):
    #Average frames by AveN to increase SNR
    import math
    im = tifffile.imread(path)
    nFrames, h,w = im.shape
    pathout = path[:-4]+'_'+str(AveN)+'ave.tif'
    nFramesOut=math.floor(nFrames/AveN)
    
    with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
        for i in tqdm.tqdm(range(nFramesOut)):
            AveFr= im[i*AveN]-im[i*AveN]
            for m in range(AveN):
                fr=im[i*AveN+m]
                AveFr=AveFr+fr/AveN
            img = AveFr.astype(np.uint8)
            tif.write(fr,continuous=True)

def apply_colormap(image,cmap='viridis'):
    """
    Function to apply false color to grayscale image
    input: 
    Image: grayscale image
    camp: colormap string. such as 'viridis'
    """
    import matplotlib.pyplot as plt
    colormap = plt.get_cmap(cmap)
    colored_image = colormap(image / 255.0)  # Normalize the image and apply colormap
    colored_image = (colored_image[:, :, :3] * 255).astype(np.uint8)  # Convert to 8-bit RGB
    return colored_image

def falsecolorstack(path,cmap='viridis',fps=15):
    import tifffile
    import numpy as np
    import cv2
    import tqdm
    print('Convert tiff stack to false color movie')
     

    stack = tifffile.imread(path)
    nFrames,h,w = stack.shape #make sure the order of h and w!
  
    pathout =path[:-4]+'_c_'+str(fps)+'.avi'    
    

    fourcc = cv2.VideoWriter_fourcc(*'MJPG')
    
    video_writer = cv2.VideoWriter(pathout, fourcc, fps, (w, h),isColor=True)
    
    # Process each frame and write to the video
    for i in tqdm.tqdm(range(nFrames)):
        grayscale_image = stack[i]
        colored_image = apply_colormap(grayscale_image,cmap)
        colored_image = cv2.cvtColor(colored_image, cv2.COLOR_RGB2BGR)  # Convert RGB to BGR
    
        video_writer.write(colored_image)
    
    # Release the video writer
    video_writer.release()
    print('Convertion done!')

def convertAVIcodec(path,fpsout=0,exp_t=0.2,nFrames=None,isColor=False):
    """
    Function to Directly convert screen recording avi to avi that mac can read
    If fpsout=0, the framerate will not be changed (same as original)
    exp_t: actual exposure time, to remove dupulicate frames
    else fpsout will be changed to set value
    """

    import moviepy.editor as mp
    print("======================")
    print("Convert AVI codec!")
    video = mp.VideoFileClip(path)
    fps = int(video.fps)
    w = int(video.w)
    h = int(video.h)
    if nFrames==None:
        nFrames = int(fps*video.duration)
        duration =video.duration
    print("fps=", fps, "   w =",w, "   h=", h)
    print("Total number of Frames=", nFrames)
    #Remove duplic frames
    if fpsout==0:
        fpsout=fps
        nFramesOut=nFrames
        dup=1
    else:
        dup=fps*exp_t
        nFramesOut=int(nFrames/dup)
  
    # Export avi
    if nFrames==None:
        pathout2 =path[:-4]+'_'+str(fpsout)+'.avi'      
        pathout= path[:-4]+'_DB.tif'    
    else:
        pathout2 =path[:-4]+'_Trim_'+str(fpsout)+'.avi'      
        pathout= path[:-4]+'_Trim_DB.tif'   
    codec  = cv2.VideoWriter_fourcc(*'MJPG')
    video_out = cv2.VideoWriter(pathout2, codec , fpsout, (w,  h),isColor=isColor)
    
    with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
        for j in tqdm.tqdm(range(nFramesOut)):
            i=j*dup
            fr = video.get_frame(mp.cvsecs(i/fps))
            if isColor==False:
                fr = cv2.cvtColor(fr, cv2.COLOR_BGR2GRAY)

            tif.write(fr, contiguous=True)
            if isColor== True:
                fr = cv2.cvtColor(fr,cv2.COLOR_BGR2RGB)#since open cv uses different color sequence
            video_out.write(fr)
        
    
    video.reader.close()
    video_out.release()
    print("Conversion done!")


def invertLUT(path):
    import tifffile
    import tqdm

    video = tifffile.imread(path)
    nFrames, h,w = video.shape[:3]
    pathout =path[:-4]+'_Inv.tiff'    

    with tifffile.TiffWriter(pathout,bigtiff=True) as tif:
        
        print("---------------------------------------------")
        print('Inverting Tiff stack LUT')    
        for i in tqdm.tqdm(range(nFrames)):        
            img=video[i]  
            img_IV=255-img
            tif.write(img_IV, contiguous=True)
        print("Invert LUT done!")










